#//
#//-----------------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2013 NVIDIA Corporation
#//   Copyright 2019-2020 Tuomas Poikela
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//-----------------------------------------------------------------------------


from .sv import sv
from ..dap.uvm_set_before_get_dap import uvm_set_before_get_dap
from .uvm_object import UVMObject
from ..macros.uvm_object_defines import uvm_object_utils
from ..macros.uvm_message_defines import uvm_error, uvm_warning
from .uvm_globals import uvm_sim_time

from .uvm_recorder import UVMTextRecorder

#//------------------------------------------------------------------------------
#// File: Transaction Recording Streams
#//
#
#// class- m_uvm_tr_stream_cfg
#// Undocumented helper class for storing stream
#// initialization values.

class m_uvm_tr_stream_cfg():
    def __init__(self):
        self.db = None  # uvm_tr_database db
        self.scope = ""
        self.stream_type_name = ""
#endclass : m_uvm_tr_stream_cfg

#
#typedef class uvm_set_before_get_dap
#typedef class uvm_text_recorder

#//------------------------------------------------------------------------------
#//
#// CLASS: UVMTrStream
#//
#// The ~UVMTrStream~ base class is a representation of a stream of records
#// within a <uvm_tr_database>.
#//
#// The record stream is intended to hide the underlying database implementation
#// from the end user, as these details are often vendor or tool-specific.
#//
#// The ~UVMTrStream~ class is pure virtual, and must be extended with an
#// implementation.  A default text-based implementation is provided via the
#// <UVMTextTrStream> class.
#//
#virtual class UVMTrStream extends uvm_object

class UVMTrStream(UVMObject):
    #
    #   // Variable- m_cfg_dap
    #   // Data access protected reference to the DB
    #   local uvm_set_before_get_dap#(m_uvm_tr_stream_cfg) m_cfg_dap
    #
    #
    #   // Variable- m_warn_null_cfg
    #   // Used to limit the number of warnings
    #   local bit m_warn_null_cfg
    #
    #   // Variable- m_is_opened
    #   // Used to indicate stream is open
    #   local bit m_is_opened
    #
    #   // Variable- m_is_closed
    #   // Used to indicate stream is closed
    #   local bit m_is_closed
    #
    #   // !m_is_opened && !m_is_closed == m_is_freed
    #

    def __init__(self, name="unnamed-UVMTrStream"):
        """
           Function: new
           Constructor

           Parameters:
           name - Stream instance name
        Args:
            name:
        """
        UVMObject.__init__(self, name)
        self.m_cfg_dap = uvm_set_before_get_dap("cfg_dap")
        #   // Variable- m_records
        #   // Active records in the stream (active == open or closed)
        self.m_records = {}  # bit [uvm_recorder]
        self.m_warn_null_cfg = 0


    #   // Variable- m_ids_by_stream
    #   // An associative array of integers, indexed by uvm_tr_streams.  This
    #   // provides a unique 'id' or 'handle' for each stream, which can be
    #   // used to identify the stream.
    #   //
    #   // By default, neither ~m_ids_by_stream~ or ~m_streams_by_id~ are
    #   // used.  Streams are only placed in the arrays when the user
    #   // attempts to determine the id for a stream.
    #   local static integer m_ids_by_stream[UVMTrStream]
    m_ids_by_stream = {}

    def get_db(self):
        """

           Group: Configuration API

           Function: get_db
           Returns a reference to the database which contains this
           stream.

           A warning will be asserted if get_db is called prior to
           the stream being initialized via `do_open`.
        Returns:
        """
        m_cfg = []  # m_uvm_tr_stream_cfg m_cfg
        if not self.m_cfg_dap.try_get(m_cfg):
            if self.m_warn_null_cfg == 1:
                uvm_warning("UVM/REC_STR/NO_CFG",
                    sv.sformatf("attempt to retrieve DB from '%s' before it was set!",
                        self.get_name()))
            self.m_warn_null_cfg = 0
            return None
        return m_cfg[0].db


    #   // Function: get_scope
    #   // Returns the ~scope~ supplied when opening this stream.
    #   //
    #   // A warning will be asserted if get_scope is called prior to
    #   // the stream being initialized via <do_open>.
    #   function string get_scope()
    #      m_uvm_tr_stream_cfg m_cfg
    #      if (!m_cfg_dap.try_get(m_cfg)) begin
    #         if (m_warn_null_cfg == 1)
    #           `uvm_warning("UVM/REC_STR/NO_CFG",
    #                        $sformatf("attempt to retrieve scope from '%s' before it was set!",
    #                                  get_name()))
    #         m_warn_null_cfg = 0
    #         return ""
    #      end
    #      return m_cfg.scope
    #   endfunction : get_scope
    #
    #   // Function: get_stream_type_name
    #   // Returns a reference to the database which contains this
    #   // stream.
    #   //
    #   // A warning will be asserted if get_stream_type_name is called prior to
    #   // the stream being initialized via <do_open>.
    #   function string get_stream_type_name()
    #      m_uvm_tr_stream_cfg m_cfg
    #      if (!m_cfg_dap.try_get(m_cfg)) begin
    #         if (m_warn_null_cfg == 1)
    #           `uvm_warning("UVM/REC_STR/NO_CFG",
    #                        $sformatf("attempt to retrieve STREAM_TYPE_NAME from '%s' before it was set!",
    #                                  get_name()))
    #         m_warn_null_cfg = 0
    #         return ""
    #      end
    #      return m_cfg.stream_type_name
    #   endfunction : get_stream_type_name
    #
    #   // Group: Stream API
    #   //
    #   // Once a stream has been opened via <uvm_tr_database::open_stream>, the user
    #   // can ~close~ the stream.
    #   //
    #   // Due to the fact that many database implementations will require crossing
    #   // a language boundary, an additional step of ~freeing~ the stream is required.
    #   //
    #   // A ~link~ can be established within the database any time between "Open" and
    #   // "Free", however it is illegal to establish a link after "Freeing" the stream.
    #   //
    #

    #   // Function: close
    #   // Closes this stream.
    #   //
    #   // Closing a stream closes all open recorders in the stream.
    #   //
    #   // This method will trigger a <do_close> call, followed by
    #   // <uvm_recorder::close> on all open recorders within the
    #   // stream.
    #   function void close()
    #      if (!is_open())
    #        return
    #
    #      do_close()
    #
    #      foreach (m_records[idx])
    #        if (idx.is_open())
    #          idx.close()
    #
    #      m_is_opened = 0
    #      m_is_closed = 1
    #   endfunction : close
    #

    #   // Function: free
    #   // Frees this stream.
    #   //
    #   // Freeing a stream indicates that the database can free any
    #   // references to the stream (including references to records
    #   // within the stream).
    #   //
    #   // This method will trigger a <do_free> call, followed by
    #   // <uvm_recorder::free> on all recorders within the stream.
    #   function void free()
    #	   process p
    #	   string s
    #      uvm_tr_database db
    #      if (!is_open() && !is_closed())
    #        return
    #
    #      if (is_open())
    #        close()
    #
    #      do_free()
    #
    #      foreach (m_records[idx])
    #        idx.free()
    #
    #      // Clear out internal state
    #      db = get_db()
    #      m_is_closed = 0
    #      p = process::self()
    #      if(p != None)
    #      	s = p.get_randstate()
    #      m_cfg_dap = new("cfg_dap")
    #      if(p != None)
    #      	p.set_randstate(s)
    #      m_warn_null_cfg = 1
    #      if (m_ids_by_stream.exists(this))
    #        m_free_id(m_ids_by_stream[this])
    #
    #      // Clear out DB state
    #      if (db != None)
    #        db.m_free_stream(this)
    #   endfunction : free

    def m_do_open(self, db, scope="", stream_type_name=""):
        """
           Function- m_do_open
           Initializes the state of the stream

           Parameters-
           db - Database which the stream belongs to
           scope - Optional scope
           stream_type_name - Optional type name for the stream

           This method will trigger a `do_open` call.

           An error will be asserted if-
           - m_do_open is called more than once without the stream
             being `freed` between.
           - m_do_open is passed a `None` db
          function void m_do_open(uvm_tr_database db,
                                  string scope="",
                                  string stream_type_name="")
        Args:
            db:
            scope:
            stream_type_name:
        """
        m_cfg = []  # m_uvm_tr_stream_cfg
        m_db = None  # uvm_tr_database
        if db is None:
            uvm_error("UVM/REC_STR/NULL_DB",
                       sv.sformatf("Illegal attempt to set DB for '%s' to '<None>'",
                                 self.get_full_name()))
            return

        if self.m_cfg_dap.try_get(m_cfg):
            uvm_error("UVM/REC_STR/RE_CFG",
                       sv.sformatf("Illegal attempt to re-open '%s'",
                                 self.get_full_name()))
        else:
            # Never set before
            m_cfg = m_uvm_tr_stream_cfg()
            m_cfg.db = db
            m_cfg.scope = scope
            m_cfg.stream_type_name = stream_type_name
            self.m_cfg_dap.set(m_cfg)
            self.m_is_opened = True

            self.do_open(db, scope, stream_type_name)


    def is_open(self):
        """
        Function: is_open
        Returns true if this `UVMTrStream` was opened on the database,
        but has not yet been closed.

        Returns:
        """
        return self.m_is_opened

    def is_closed(self):
        """
           Function: is_closed
           Returns true if this `UVMTrStream` was closed on the database,
           but has not yet been freed.

        Returns:
        """
        return self.m_is_closed

    #   // Group: Transaction Recorder API
    #   //
    #   // New recorders can be opened prior to the stream being ~closed~.
    #   //
    #   // Once a stream has been closed, requests to open a new recorder
    #   // will be ignored (<open_recorder> will return ~None~).
    #   //
    #

    def open_recorder(self, name, open_time=0, type_name=""):
        """
           Function: open_recorder
           Marks the opening of a new transaction recorder on the stream.

           Parameters:
           name - A name for the new transaction
           open_time - Optional time to record as the opening of this transaction
           type_name - Optional type name for the transaction

           If `open_time` is omitted (or set to 0), then the stream will use
           the current time.

           This method will trigger a `do_open_recorder` call.  If `do_open_recorder`
           returns a non-`None` value, then the <uvm_recorder::do_open> method will
           be called in the recorder.

           Transaction recorders can only be opened if the stream is
           `open` on the database (per `is_open`).  Otherwise the
           request will be ignored, and `None` will be returned.
        Args:
            name:
            open_time:
            type_name:
        Returns:
        """
        m_time = open_time
        if open_time == 0:
            m_time = sv.time()

        # Check to make sure we're open
        if not self.is_open():
            return None
        else:
            #process p = process::self()
            p = None
            s = ""
            if p is not None:
                s = p.get_randstate()
            open_recorder = self.do_open_recorder(name, m_time, type_name)

            if open_recorder is not None:
                self.m_records[open_recorder] = 1
                open_recorder.m_do_open(self, m_time, type_name)

            if p is not None:
                p.set_randstate(s)
            return open_recorder
        #   endfunction : open_recorder


    def m_free_recorder(self, recorder):
        """
           Function- m_free_recorder
           Removes recorder from the internal array
        Args:
            recorder:
        """
        if recorder in self.m_records:
            del self.m_records[recorder]

    #
    #   // Function: get_recorders
    #   // Provides a queue of all transactions within the stream.
    #   //
    #   // Parameters:
    #   // q - A reference to the queue of <uvm_recorder>s
    #   //
    #   // The <get_recorders> method returns the size of the queue,
    #   // such that the user can conditionally process the elements.
    #   //
    #   // | uvm_recorder tr_q[$]
    #   // | if (my_stream.get_recorders(tr_q)) begin
    #   // |   // Process the queue...
    #   // | end
    #   //
    #   function unsigned get_recorders(ref uvm_recorder q[$])
    #      // Clear out the queue first...
    #      q.delete()
    #      // Fill in the values
    #      foreach (m_records[idx])
    #        q.push_back(idx)
    #      // Finally return the size of the queue
    #      return q.size()
    #   endfunction : get_recorders
    #

    #   // Group: Handles
    #
    #   // Variable- m_streams_by_id
    #   // A corollary to ~m_ids_by_stream~, this indexes the streams by their
    #   // unique ids.
    #   local static UVMTrStream m_streams_by_id[integer]
    m_streams_by_id = {}


    def get_handle(self):
        """
           Function: get_handle
           Returns a unique ID for this stream.

           A value of `0` indicates that the recorder has been `freed`,
           and no longer has a valid ID.

        Returns:
        """
        m_streams_by_id = UVMTrStream.m_streams_by_id
        m_ids_by_stream = UVMTrStream.m_ids_by_stream

        if self.is_open() is False and self.is_closed() is False:
            return 0
        else:
            handle = self.get_inst_id()

            # Check for the weird case where our handle changed.
            if self in m_ids_by_stream and m_ids_by_stream[self] != handle:
                key = m_ids_by_stream[self]
                del m_streams_by_id[key]

            m_streams_by_id[handle] = self
            m_ids_by_stream[self] = handle

            return handle
    #   endfunction : get_handle

    #
    #   // Function- m_get_handle
    #   // Provided to allow implementation-specific handles which are not
    #   // identical to the built-in handles.
    #   //
    #   // This is an implementation detail of the UVM library, which allows
    #   // for vendors to (optionally) put vendor-specific methods into the library.
    #   virtual function integer m_get_handle()
    #      return get_handle()
    #   endfunction : m_get_handle
    #

    #   // Function: get_stream_from_handle
    #   // Static accessor, returns a stream reference for a given unique id.
    #   //
    #   // If no stream exists with the given ~id~, or if the
    #   // stream with that ~id~ has been freed, then ~None~ is
    #   // returned.
    #   //
    #   static function UVMTrStream get_stream_from_handle(integer id)
    #      if (id == 0)
    #        return None
    #
    #      if ($isunknown(id) || !m_streams_by_id.exists(id))
    #        return None
    #
    #      return m_streams_by_id[id]
    #   endfunction : get_stream_from_handle
    #

    #   // Function- m_free_id
    #   // Frees the id/stream link (memory cleanup)
    #   //
    #   static function void m_free_id(integer id)
    #      UVMTrStream stream
    #      if (!$isunknown(id) && m_streams_by_id.exists(id))
    #        stream = m_streams_by_id[id]
    #
    #      if (stream != None) begin
    #         m_streams_by_id.delete(id)
    #         m_ids_by_stream.delete(stream)
    #      end
    #   endfunction : m_free_id
    #
    #   // Group: Implementation Agnostic API
    #   //
    #

    #   // Function: do_open
    #   // Callback triggered via <uvm_tr_database::open_stream>.
    #   //
    #   // Parameters:
    #   // db - Database which the stream belongs to
    #   // scope - Optional scope
    #   // stream_type_name - Optional type name for the stream
    #   //
    #   // The ~do_open~ callback can be used to initialize any internal
    #   // state within the stream, as well as providing a location to
    #   // record any initial information about the stream.
    #   protected virtual function void do_open(uvm_tr_database db,
    #                                           string scope,
    #                                           string stream_type_name)
    #   endfunction : do_open

    #
    #   // Function: do_close
    #   // Callback triggered via <close>.
    #   //
    #   // The ~do_close~ callback can be used to set internal state
    #   // within the stream, as well as providing a location to
    #   // record any closing information.
    #   protected virtual function void do_close()
    #   endfunction : do_close
    #

    #   // Function: do_free
    #   // Callback triggered via <free>.
    #   //
    #   // The ~do_free~ callback can be used to release the internal
    #   // state within the stream, as well as providing a location
    #   // to record any "freeing" information.
    #   protected virtual function void do_free()
    #   endfunction : do_free
    #

    #   // Function: do_open_recorder
    #   // Marks the beginning of a new record in the stream.
    #   //
    #   // Backend implementation of <open_recorder>
    #   protected virtual function uvm_recorder do_open_recorder(string name,
    #                                                            time   open_time,
    #                                                            string type_name)
    #      return None
    #   endfunction : do_open_recorder
    #
    #endclass : UVMTrStream
#

#//------------------------------------------------------------------------------
#//
#// CLASS: UVMTextTrStream
#//
#// The ~UVMTextTrStream~ is the default stream implementation for the
#// <uvm_text_tr_database>.
#//
#//
#
#class UVMTextTrStream extends UVMTrStream
class UVMTextTrStream(UVMTrStream):

    def __init__(self, name="unnamed-UVMTextTrStream"):
        UVMTrStream.__init__(self, name)
        #   // Variable- m_text_db
        #   // Internal reference to the text-based backend
        self.m_text_db = None


    #   // Group: Implementation Agnostic API

    def do_open(self, db, scope, stream_type_name):
        """
           Function: do_open
           Callback triggered via <uvm_tr_database::open_stream>.

          protected virtual function void do_open(uvm_tr_database db,
                                                  string scope,
                                                  string stream_type_name)
        Args:
            db:
            scope:
            stream_type_name:
        """
        # $cast(m_text_db, db)
        self.m_text_db = db
        if self.m_text_db.open_db():
            sv.fdisplay(self.m_text_db.m_file,
                      "  CREATE_STREAM @%0t [NAME:%s T:%s SCOPE:%s STREAM:%0d]",
                      uvm_sim_time(),
                      self.get_name(),
                      stream_type_name,
                      scope,
                      self.get_handle())

    #   // Function: do_close
    #   // Callback triggered via <UVMTrStream::close>.
    #   protected virtual function void do_close()
    #      if (m_text_db.open_db())
    #        $fdisplay(m_text_db.m_file,
    #                  "  CLOSE_STREAM @%0t {NAME:%s T:%s SCOPE:%s STREAM:%0d}",
    #                  $time,
    #                  this.get_name(),
    #                  this.get_stream_type_name(),
    #                  this.get_scope(),
    #                  this.get_handle())
    #   endfunction : do_close
    #

    #   // Function: do_free
    #   // Callback triggered via <UVMTrStream::free>.
    #   //
    #   protected virtual function void do_free()
    #      if (m_text_db.open_db())
    #        $fdisplay(m_text_db.m_file,
    #                  "  FREE_STREAM @%0t {NAME:%s T:%s SCOPE:%s STREAM:%0d}",
    #                  $time,
    #                  this.get_name(),
    #                  this.get_stream_type_name(),
    #                  this.get_scope(),
    #                  this.get_handle())
    #      m_text_db = None
    #      return
    #   endfunction : do_free
    #

    def do_open_recorder(self, name, open_time, type_name):
        """
           Function: do_open_recorder
           Marks the beginning of a new record in the stream

           Text-backend specific implementation.
        Args:
            name:
            open_time:
            type_name:
        Returns:
        """
        if self.m_text_db.open_db():
            return UVMTextRecorder.type_id.create(name)
        return None


    #endclass : UVMTextTrStream
uvm_object_utils(UVMTextTrStream)



#//
#//-----------------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2019 Tuomas Poikela
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//-----------------------------------------------------------------------------

from ..base.uvm_component import UVMComponent


class UVMScoreboard(UVMComponent):
    """
    The `UVMScoreboard` class should be used as the base class for
    user-defined scoreboards.

    Deriving from `UVMScoreboard` will allow you to distinguish scoreboards from
    other component types inheriting directly from `UVMComponent`. Such
    scoreboards will automatically inherit and benefit from features that may be
    added to `UVMScoreboard` in the future.
    """

    #  // Function: new
    #  //
    #  // Creates and initializes an instance of this class using the normal
    #  // constructor arguments for `UVMComponent`: ~name~ is the name of the
    #  // instance, and ~parent~ is the handle to the hierarchical parent, if any.
    def __init__(self, name, parent):
        UVMComponent.__init__(self, name, parent)

    type_name = "uvm_scoreboard"

    def get_type_name(self):
        return UVMScoreboard.type_name



#//
#//------------------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2019 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//------------------------------------------------------------------------------

from typing import TypeVar, Dict, Generic, List, Optional, Any
import collections

from .uvm_object import UVMObject
from ..macros import uvm_warning

KEY = TypeVar('KEY')
VAL = TypeVar('VAL')

class UVMPool(UVMObject, Generic[KEY, VAL]):
    """
    Implements a class-based dynamic associative array. Allows sparse arrays to
    be allocated on demand, and passed and stored by reference.
    """

    type_name = "uvm_pool"
    m_global_pool = None


    def __init__(self, name="", T=None):
        UVMObject.__init__(self, name)
        self.pool = collections.OrderedDict()
        self.ptr = -1
        self.T = T

    @classmethod
    def get_global_pool(cls) -> 'UVMPool':
        """
        Returns the singleton global pool. Unlike in SV, uvm-python has only one
        global pool.

        This allows items to be shared amongst components throughout the
        verification environment.

        Returns:
            UVMPool: Single global pool
        """
        if UVMPool.m_global_pool is None:
            UVMPool.m_global_pool = UVMPool("pool")
        return UVMPool.m_global_pool

    @classmethod
    def get_global(cls, key) -> Any:
        """
        Returns the specified item instance from the global item pool.

        Args:
            key (str):
        Returns:
            any: Item matching the given key.
        """
        gpool = UVMPool.get_global_pool()
        return gpool.get(key)

    def get(self, key: KEY) -> VAL:
        """
        Function: get

        Returns the item with the given `key`.

        If no item exists by that key, a new item is created with that key
        and returned.
        Args:
            key:
        Returns:
        """
        if key in self.pool:
            return self.pool[key]
        elif self.T is not None:
            self.pool[key] = self.T()
            return self.pool[key]
        return None

    def add(self, key: KEY, item: VAL) -> None:
        self.pool[key] = item

    def num(self) -> int:
        return len(self.pool.keys())

    def keys(self):
        return self.pool.keys()

    def key_list(self) -> List[KEY]:
        return list(self.pool.keys())

    def delete(self, key=None) -> None:
        if key is None:
            self.pool = {}
        else:
            if key in self.pool:
                del self.pool[key]

    def exists(self, key: KEY) -> bool:
        return key in self.pool

    def last(self):
        keys = self.pool.keys()
        if len(keys) > 0:
            self.ptr = self.num() - 1
            return next(reversed(self.pool))  # type: ignore
        else:
            return False

    def has_first(self) -> bool:
        return self.num() > 0

    def has_last(self) -> bool:
        return self.num() > 0

    def first(self):
        # keys = self.pool.keys()
        for k in self.pool:
            self.ptr = 0
            return k
        return False

    def has_next(self) -> bool:
        if self.ptr < (self.num() - 1):
            return True
        return False

    def next(self) -> Optional[KEY]:
        if self.has_next() is True:
            self.ptr += 1
            key = list(self.pool.keys())[self.ptr]
            return key
        return None

    def has_prev(self) -> bool:
        if self.ptr > 0:
            return True
        return False

    def prev(self) -> Optional[KEY]:
        if self.has_prev() is True:
            self.ptr -= 1
            key = list(self.pool.keys())[self.ptr]
            return key
        return None

    def create(self, name="") -> 'UVMPool':
        return UVMPool(name, self.T)

    def do_print(self, printer) -> None:
        while self.has_next():
            key = self.next()
            item = self.pool[key]
            # print_generic(self, name, type_name, size, value, scope_separator="."):
            if hasattr(item, 'convert2string'):
                printer.print_string(item.get_name(), item.convert2string())
            else:
                name = ""
                if hasattr(key, 'get_name'):
                    name = key.get_name()
                printer.print_generic(name, '', 0, str(item))

    def __len__(self):
        """
        len() operator
        Returns:
        """
        return self.num()

    def __contains__(self, key) -> bool:
        """
        Implements X in Y operator
        Args:
            key:
        Returns:
        """
        return key in self.pool

    def __setitem__(self, key, value):
        """
        Implements aa[x] = y
        Args:
            key:
            value:
        """
        self.pool[key] = value

    def __getitem__(self, key):
        """
        Implements aa[x]
        Args:
            key:
        Returns:
        Raises:
        """
        if key in self.pool:
            return self.pool[key]
        else:
            raise IndexError('No key found')




class UVMObjectStringPool(UVMPool[str, UVMObject]):
    """
    This provides a specialization of the generic `UVMPool` class for
    an associative array of `UVMObject`-based objects indexed by string.
    Specializations of this class include the `UVMEventPool` (a
    `UVMObjectStringPool` storing `UVMEvent` and
    `UVMBarrierPool` (a `UVMObjectStringPool` storing `UVMBarrier`).
    """

    m_global_pool = None
    type_name = "uvm_obj_str_pool"

    def __init__(self, name="", Constr=UVMObject):
        """
        Creates a new pool with the given `name`.

        Args:
            name (str): Name of the pool.
            Constr (class): Used for constructing default objects.
        """
        UVMPool.__init__(self, name)
        self.Constructor = Constr


    def get_type_name(self) -> str:
        """
        Returns the type name of this object.

        Returns:
            str: Type name of this object.
        """
        return UVMObjectStringPool.type_name


    @classmethod
    def get_global_pool(cls) -> 'UVMObjectStringPool':
        """
        Returns the singleton global pool.

        This allows objects to be shared amongst components throughout the
        verification environment.

        Returns:
            UVMObjectStringPool: Global pool
        """
        if UVMObjectStringPool.m_global_pool is None:
            UVMObjectStringPool.m_global_pool = UVMObjectStringPool("global_pool")
        return UVMObjectStringPool.m_global_pool


    @classmethod
    def get_global(cls, key: str) -> UVMObject:
        """
        Returns the specified item instance from the global item pool.

        Args:
            key (str): Key used for getting the item from global pool.

        Returns:
            any: Object matching the key in the global pool.
        """
        gpool = UVMObjectStringPool.get_global_pool()
        return gpool.get(key)


    def get(self, key: str) -> UVMObject:
        """
        Returns the object item at the given string `key`.

        If no item exists by the given `key`, a new item is created for that key
        and returned.

        Args:
            key (str): Key used for getting the item.
        Returns:
            any: Item matching the key. If no match, returns new object.
        """
        if key not in self.pool:
            self.pool[key] = self.Constructor(key)
        return self.pool[key]


    def delete(self, key: str) -> None:
        """
        Removes the item with the given string `key` from the pool.

        Args:
            key (str): Key used for removing the item.
        """
        if not self.exists(key):
            uvm_warning("POOLDEL", "delete: key '{}' doesn't exist".format(key))
            return
        self.delete(key)
    #  endfunction

    def do_print(self, printer) -> None:
        """
          Function- do_print

        Args:
            printer (UVMPrinter): Printer used for printing
        """
        key = ""
        num_keys = len(list(self.pool.keys()))
        printer.print_array_header("pool", num_keys,"aa_object_string")
        if self.has_first():
            key = self.first()
            while True:
                printer.print_object("[" + key + "]", self.pool[key],"[")
                if self.has_next():
                    key = self.next()
                else:
                    break
        printer.print_array_footer()
    # endfunction


class UVMEventPool(UVMObjectStringPool):
    def __init__(self, name=""):
        from .uvm_event import UVMEvent
        UVMObjectStringPool.__init__(self, name, UVMEvent)


class UVMBarrierPool(UVMObjectStringPool):
    def __init__(self, name=""):
        from .uvm_barrier import UVMBarrier
        UVMObjectStringPool.__init__(self, name, UVMBarrier)



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2009 Synopsys, Inc.
#//    Copyright 2010 Mentor Graphics Corporation
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//
"""
#//------------------------------------------------------------------------------
#// Title: Virtual Register Field Classes
#//
#// This section defines the virtual field and callback classes.
#//
#// A virtual field is set of contiguous bits in one or more memory locations.
#// The semantics and layout of virtual fields comes from
#// an agreement between the software and the hardware,
#// not any physical structures in the DUT.
#//
#//------------------------------------------------------------------------------
"""

#typedef class uvm_vreg_field_cbs

#import cocotb

from ..base import (UVMObject, sv)
from ..macros import (uvm_error, uvm_object_utils, UVM_REG_DATA_WIDTH)
from ..base.uvm_callback import *
#from .uvm_reg_model import (UVM_REG_DATA_WIDTH)


class UVMVRegField(UVMObject):
    """
     Class: UVMVRegField

     Virtual field abstraction class

     A virtual field represents a set of adjacent bits that are
     logically implemented in consecutive memory locations.
    """


    #   `uvm_register_cb(UVMVRegField, uvm_vreg_field_cbs)

    #   //
    #   // Function: new
    #   // Create a new virtual field instance
    #   //
    #   // This method should not be used directly.
    #   // The UVMVRegField::type_id::create() method should be used instead.
    #   //
    def __init__(self, name="UVMVRegField"):
        super().__init__(name)
        self.parent = None  # UVMVReg
        self.lsb = 0
        self.size = 0
        self.fname = 0
        self.lineno = 0
        self.read_in_progress = False
        self.write_in_progress = False


    #   //
    #   // Function: configure
    #   // Instance-specific configuration
    #   //
    #   // Specify the ~parent~ virtual register of this virtual field, its
    #   // ~size~ in bits, and the position of its least-significant bit
    #   // within the virtual register relative to the least-significant bit
    #   // of the virtual register.
    #   //
    #   extern function void configure(uvm_vreg parent,
    #                                  int unsigned size,
    #                                  int unsigned lsb_pos)
    def configure(self, parent, size: int, lsb_pos: int):
        self.parent = parent
        if size == 0:
            uvm_error("RegModel", sv.sformatf("Virtual field '%s' cannot have 0 bits", self.get_full_name()))
            size = 1

        if size > UVM_REG_DATA_WIDTH:
            uvm_error("RegModel", sv.sformatf("Virtual field '%s' cannot have more than %0d bits",
                self.get_full_name(),
                UVM_REG_DATA_WIDTH))
            size = UVM_REG_DATA_WIDTH

        self.size = size
        self.lsb = lsb_pos
        self.parent.add_field(self)


    #   //
    #   // Group: Introspection
    #   //

    #   //
    #   // Function: get_name
    #   // Get the simple name
    #   //
    #   // Return the simple object name of this virtual field
    #   //

    #   // Function: get_full_name
    #   // Get the hierarchical name
    #   //
    #   // Return the hierarchal name of this virtual field
    #   // The base of the hierarchical name is the root block.
    #   //
    #   extern def string        get_full_name(self):
    def get_full_name(self) -> str:
        return (self.parent.get_full_name() + "." + self.get_name())


    #   //
    #   // FUNCTION: get_parent
    #   // Get the parent virtual register
    #   //
    #   extern def get_parent(self):
    def get_parent(self):
        return self.parent

    #   extern def get_register(self):
    def get_register(self):
        return self.parent

    #   //
    #   // FUNCTION: get_lsb_pos_in_register
    #   // Return the position of the virtual field
    #   ///
    #   // Returns the index of the least significant bit of the virtual field
    #   // in the virtual register that instantiates it.
    #   // An offset of 0 indicates a field that is aligned with the
    #   // least-significant bit of the register.
    #   //
    #   extern def int unsigned get_lsb_pos_in_register(self):
    def get_lsb_pos_in_register(self):
        return self.lsb

    #   //
    #   // FUNCTION: get_n_bits
    #   // Returns the width, in bits, of the virtual field.
    #   //
    #   extern def int unsigned get_n_bits(self):
    def get_n_bits(self):
        return self.size

    #   //
    #   // FUNCTION: get_access
    #   // Returns the access policy of the virtual field register
    #   // when written and read via an address map.
    #   //
    #   // If the memory implementing the virtual field
    #   // is mapped in more than one address map,
    #   // an address ~map~ must be specified.
    #   // If access restrictions are present when accessing a memory
    #   // through the specified address map, the access mode returned
    #   // takes the access restrictions into account.
    #   // For example, a read-write memory accessed
    #   // through an address map with read-only restrictions would return "RO".
    #   //
    #   extern def get_access(self,uvm_reg_map map = None):


    #   //
    #   // Group: HDL Access
    #   //

    #   //
    #   // TASK: write
    #   // Write the specified value in a virtual field
    #   //
    #   // Write ~value~ in the DUT memory location(s) that implements
    #   // the virtual field that corresponds to this
    #   // abstraction class instance using the specified access
    #   // ~path~.
    #   //
    #   // If the memory implementing the virtual register array
    #   // containing this virtual field
    #   // is mapped in more than one address map,
    #   // an address ~map~ must be
    #   // specified if a physical access is used (front-door access).
    #   //
    #   // The operation is eventually mapped into
    #   // memory read-modify-write operations at the location
    #   // where the virtual register
    #   // specified by ~idx~ in the virtual register array is implemented.
    #   // If a backdoor is available for the memory implementing the
    #   // virtual field, it will be used for the memory-read operation.
    #   //
    #   extern def write(self,input  longint unsigned   idx,
    #                             output uvm_status_e  status,
    #                             input  uvm_reg_data_t     value,
    #                             input  uvm_path_e    path = UVM_DEFAULT_PATH,
    #                             input  uvm_reg_map        map = None,
    #                             input  uvm_sequence_base  parent = None,
    #                             input  uvm_object         extension = None,
    #                             input  string             fname = "",
    #                             input  int                lineno = 0)

    #   //
    #   // TASK: read
    #   // Read the current value from a virtual field
    #   //
    #   // Read from the DUT memory location(s) that implements
    #   // the virtual field that corresponds to this
    #   // abstraction class instance using the specified access
    #   // ~path~, and return the readback ~value~.
    #   //
    #   // If the memory implementing the virtual register array
    #   // containing this virtual field
    #   // is mapped in more than one address map,
    #   // an address ~map~ must be
    #   // specified if a physical access is used (front-door access).
    #   //
    #   // The operation is eventually mapped into
    #   // memory read operations at the location(s)
    #   // where the virtual register
    #   // specified by ~idx~ in the virtual register array is implemented.
    #   //
    #   extern def read(self,input  longint unsigned    idx,
    #                            output uvm_status_e   status,
    #                            output uvm_reg_data_t      value,
    #                            input  uvm_path_e     path = UVM_DEFAULT_PATH,
    #                            input  uvm_reg_map         map = None,
    #                            input  uvm_sequence_base   parent = None,
    #                            input  uvm_object          extension = None,
    #                            input  string              fname = "",
    #                            input  int                 lineno = 0)


    #   //
    #   // TASK: poke
    #   // Deposit the specified value in a virtual field
    #   //
    #   // Deposit ~value~ in the DUT memory location(s) that implements
    #   // the virtual field that corresponds to this
    #   // abstraction class instance using the specified access
    #   // ~path~.
    #   //
    #   // The operation is eventually mapped into
    #   // memory peek-modify-poke operations at the location
    #   // where the virtual register
    #   // specified by ~idx~ in the virtual register array is implemented.
    #   //
    #   extern def poke(self,input  longint unsigned    idx,
    #                            output uvm_status_e   status,
    #                            input  uvm_reg_data_t      value,
    #                            input  uvm_sequence_base   parent = None,
    #                            input  uvm_object          extension = None,
    #                            input  string              fname = "",
    #                            input  int                 lineno = 0)

    #   //
    #   // TASK: peek
    #   // Sample the current value from a virtual field
    #   //
    #   // Sample from the DUT memory location(s) that implements
    #   // the virtual field that corresponds to this
    #   // abstraction class instance using the specified access
    #   // ~path~, and return the readback ~value~.
    #   //
    #   // If the memory implementing the virtual register array
    #   // containing this virtual field
    #   // is mapped in more than one address map,
    #   // an address ~map~ must be
    #   // specified if a physical access is used (front-door access).
    #   //
    #   // The operation is eventually mapped into
    #   // memory peek operations at the location(s)
    #   // where the virtual register
    #   // specified by ~idx~ in the virtual register array is implemented.
    #   //
    #   extern def peek(self,input  longint unsigned    idx,
    #                            output uvm_status_e   status,
    #                            output uvm_reg_data_t      value,
    #                            input  uvm_sequence_base   parent = None,
    #                            input  uvm_object          extension = None,
    #                            input  string              fname = "",
    #                            input  int                 lineno = 0)

    #   //
    #   // Group: Callbacks
    #   //


    #   //
    #   // TASK: pre_write
    #   // Called before virtual field write.
    #   //
    #   // If the specified data value, access ~path~ or address ~map~ are modified,
    #   // the updated data value, access path or address map will be used
    #   // to perform the virtual register operation.
    #   //
    #   // The virtual field callback methods are invoked before the callback methods
    #   // on the containing virtual register.
    #   // The registered callback methods are invoked after the invocation
    #   // of this method.
    #   // The pre-write virtual register and field callbacks are executed
    #   // before the corresponding pre-write memory callbacks
    #   //
    #async def pre_write(self,longint unsigned     idx,
    #                          ref uvm_reg_data_t   wdat,
    #                          ref uvm_path_e  path,
    #                          ref uvm_reg_map   map)
    #   endtask: pre_write

    #   //
    #   // TASK: post_write
    #   // Called after virtual field write
    #   //
    #   // If the specified ~status~ is modified,
    #   // the updated status will be
    #   // returned by the virtual register operation.
    #   //
    #   // The virtual field callback methods are invoked after the callback methods
    #   // on the containing virtual register.
    #   // The registered callback methods are invoked before the invocation
    #   // of this method.
    #   // The post-write virtual register and field callbacks are executed
    #   // after the corresponding post-write memory callbacks
    #   //
    #async   def post_write(self,longint unsigned       idx,
    #                           uvm_reg_data_t         wdat,
    #                           uvm_path_e        path,
    #                           uvm_reg_map         map,
    #                           ref uvm_status_e  status)
    #   endtask: post_write

    #   //
    #   // TASK: pre_read
    #   // Called before virtual field read.
    #   //
    #   // If the specified access ~path~ or address ~map~ are modified,
    #   // the updated access path or address map will be used to perform
    #   // the virtual register operation.
    #   //
    #   // The virtual field callback methods are invoked after the callback methods
    #   // on the containing virtual register.
    #   // The registered callback methods are invoked after the invocation
    #   // of this method.
    #   // The pre-read virtual register and field callbacks are executed
    #   // before the corresponding pre-read memory callbacks
    #   //
    #async   def pre_read(self,longint unsigned      idx,
    #                         ref uvm_path_e   path,
    #                         ref uvm_reg_map    map)
    #   endtask: pre_read

    #   //
    #   // TASK: post_read
    #   // Called after virtual field read.
    #   //
    #   // If the specified readback data ~rdat~ or ~status~ is modified,
    #   // the updated readback data or status will be
    #   // returned by the virtual register operation.
    #   //
    #   // The virtual field callback methods are invoked after the callback methods
    #   // on the containing virtual register.
    #   // The registered callback methods are invoked before the invocation
    #   // of this method.
    #   // The post-read virtual register and field callbacks are executed
    #   // after the corresponding post-read memory callbacks
    #   //
    #async   def post_read(self,longint unsigned       idx,
    #                          ref uvm_reg_data_t     rdat,
    #                          uvm_path_e        path,
    #                          uvm_reg_map         map,
    #                          ref uvm_status_e  status)
    #   endtask: post_read


    #   extern def do_print(self,uvm_printer printer):
    #   extern virtual function string convert2string
    #   extern def clone(self):
    #   extern def do_copy(self,uvm_object rhs):
    #   extern virtual function bit do_compare (uvm_object  rhs,
    #                                          uvm_comparer comparer)
    #   extern def do_pack(self,uvm_packer packer):
    #   extern def do_unpack(self,uvm_packer packer):


uvm_object_utils(UVMVRegField)


#//------------------------------------------------------------------------------
#// Class: uvm_vreg_field_cbs
#//
#// Pre/post read/write callback facade class
#//
#//------------------------------------------------------------------------------

#class uvm_vreg_field_cbs(uvm_callback):
    #   string fname
    #   int    lineno
    #
    #   def __init__(self, name = "uvm_vreg_field_cbs")
    #      super().__init__(name)
    #   endfunction
    #
    #
    #   //
    #   // Task: pre_write
    #   // Callback called before a write operation.
    #   //
    #   // The registered callback methods are invoked before the invocation
    #   // of the virtual register pre-write callbacks and
    #   // after the invocation of the <UVMVRegField::pre_write()> method.
    #   //
    #   // The written value ~wdat~, access ~path~ and address ~map~,
    #   // if modified, modifies the actual value, access path or address map
    #   // used in the register operation.
    #   //
    #@cocotb.coroutine
    #   def pre_write(self,UVMVRegField       field,
    #                          longint unsigned     idx,
    #                          ref uvm_reg_data_t   wdat,
    #                          ref uvm_path_e  path,
    #                          ref uvm_reg_map   map)
    #   endtask: pre_write
    #
    #
    #   //
    #   // TASK: post_write
    #   // Called after a write operation
    #   //
    #   // The registered callback methods are invoked after the invocation
    #   // of the virtual register post-write callbacks and
    #   // before the invocation of the <UVMVRegField::post_write()> method.
    #   //
    #   // The ~status~ of the operation,
    #   // if modified, modifies the actual returned status.
    #   //
    #@cocotb.coroutine
    #   def post_write(self,UVMVRegField        field,
    #                           longint unsigned      idx,
    #                           uvm_reg_data_t        wdat,
    #                           uvm_path_e       path,
    #                           uvm_reg_map        map,
    #                           ref uvm_status_e status)
    #   endtask: post_write
    #
    #
    #   //
    #   // TASK: pre_read
    #   // Called before a virtual field read.
    #   //
    #   // The registered callback methods are invoked after the invocation
    #   // of the virtual register pre-read callbacks and
    #   // after the invocation of the <UVMVRegField::pre_read()> method.
    #   //
    #   // The access ~path~ and address ~map~,
    #   // if modified, modifies the actual access path or address map
    #   // used in the register operation.
    #   //
    #@cocotb.coroutine
    #   def pre_read(self,UVMVRegField        field,
    #                         longint unsigned      idx,
    #                         ref uvm_path_e   path,
    #                         ref uvm_reg_map    map)
    #   endtask: pre_read
    #
    #
    #   //
    #   // TASK: post_read
    #   // Called after a virtual field read.
    #   //
    #   // The registered callback methods are invoked after the invocation
    #   // of the virtual register post-read callbacks and
    #   // before the invocation of the <UVMVRegField::post_read()> method.
    #   //
    #   // The readback value ~rdat~ and the ~status~ of the operation,
    #   // if modified, modifies the actual returned readback value and status.
    #   //
    #@cocotb.coroutine
    #   def post_read(self,UVMVRegField         field,
    #                          longint unsigned       idx,
    #                          ref uvm_reg_data_t     rdat,
    #                          uvm_path_e        path,
    #                          uvm_reg_map         map,
    #                          ref uvm_status_e  status)
    #   endtask: post_read
    #endclass: uvm_vreg_field_cbs
#
#
#//
#// Type: uvm_vreg_field_cb
#// Convenience callback type declaration
#//
#// Use this declaration to register virtual field callbacks rather than
#// the more verbose parameterized class
#//
#typedef uvm_callbacks#(UVMVRegField, uvm_vreg_field_cbs) uvm_vreg_field_cb
#
#//
#// Type: uvm_vreg_field_cb_iter
#// Convenience callback iterator type declaration
#//
#// Use this declaration to iterate over registered virtual field callbacks
#// rather than the more verbose parameterized class
#//
#typedef uvm_callback_iter#(UVMVRegField, uvm_vreg_field_cbs) uvm_vreg_field_cb_iter
#
#
#
#def string UVMVRegField::get_access(self,uvm_reg_map map = None):
#   if (self.parent.get_memory() is None):
#      uvm_error("RegModel", sv.sformatf("Cannot call UVMVRegField::get_rights() on unimplemented virtual field \"%s\"",
#                                     self.get_full_name()))
#      return "RW"
#   end
#
#   return self.parent.get_access(map)
#endfunction: get_access
#
#
#@cocotb.coroutine
#task UVMVRegField::write(input  longint unsigned    idx,
#                           output uvm_status_e   status,
#                           input  uvm_reg_data_t      value,
#                           input  uvm_path_e     path = UVM_DEFAULT_PATH,
#                           input  uvm_reg_map      map = None,
#                           input  uvm_sequence_base   parent = None,
#                           input  uvm_object          extension = None,
#                           input  string              fname = "",
#                           input  int                 lineno = 0)
#   uvm_reg_data_t  tmp
#   uvm_reg_data_t  segval
#   uvm_reg_addr_t  segoff
#   uvm_status_e st
#
#   int flsb, fmsb, rmwbits
#   int segsiz, segn
#   uvm_mem    mem
#   uvm_path_e rm_path
#
#   uvm_vreg_field_cb_iter cbs = new(self)
#
#   self.fname = fname
#   self.lineno = lineno
#
#   write_in_progress = 1'b1
#   mem = self.parent.get_memory()
#   if (mem is None):
#      uvm_error("RegModel", sv.sformatf("Cannot call UVMVRegField::write() on unimplemented virtual register \"%s\"",
#                                     self.get_full_name()))
#      status = UVM_NOT_OK
#      return
#   end
#
#   if (path == UVM_DEFAULT_PATH):
#      uvm_reg_block blk = self.parent.get_block()
#      path = blk.get_default_path()
#   end
#
#   status = UVM_IS_OK
#
#   self.parent.XatomicX(1)
#
#   if (value >> self.size):
#      uvm_warning("RegModel", sv.sformatf("Writing value 'h%h that is greater than field \"%s\" size (%0d bits)", value, self.get_full_name(), self.get_n_bits()))
#      value &= value & ((1<<self.size)-1)
#   end
#   tmp = 0
#
#   self.pre_write(idx, value, path, map)
#   for (uvm_vreg_field_cbs cb = cbs.first(); cb is not None
#        cb = cbs.next()):
#      cb.fname = self.fname
#      cb.lineno = self.lineno
#      cb.pre_write(self, idx, value, path, map)
#   end
#
#   segsiz = mem.get_n_bytes() * 8
#   flsb    = self.get_lsb_pos_in_register()
#   segoff  = self.parent.get_offset_in_memory(idx) + (flsb / segsiz)
#
#   // Favor backdoor read to frontdoor read for the RMW operation
#   rm_path = UVM_DEFAULT_PATH
#   if (mem.get_backdoor() is not None) rm_path = UVM_BACKDOOR
#
#   // Any bits on the LSB side we need to RMW?
#   rmwbits = flsb % segsiz
#
#   // Total number of memory segment in this field
#   segn = (rmwbits + self.get_n_bits() - 1) / segsiz + 1
#
#   if (rmwbits > 0):
#      uvm_reg_addr_t  segn
#
#      mem.read(st, segoff, tmp, rm_path, map, parent, , extension, fname, lineno)
#      if (st != UVM_IS_OK  and  st != UVM_HAS_X):
#         uvm_error("RegModel",
#                    sv.sformatf("Unable to read LSB bits in %s[%0d] to for RMW cycle on virtual field %s.",
#                              mem.get_full_name(), segoff, self.get_full_name()))
#         status = UVM_NOT_OK
#         self.parent.XatomicX(0)
#         return
#      end
#
#      value = (value << rmwbits) | (tmp & ((1<<rmwbits)-1))
#   end
#
#   // Any bits on the MSB side we need to RMW?
#   fmsb = rmwbits + self.get_n_bits() - 1
#   rmwbits = (fmsb+1) % segsiz
#   if (rmwbits > 0):
#      if (segn > 0):
#         mem.read(st, segoff + segn - 1, tmp, rm_path, map, parent,, extension, fname, lineno)
#         if (st != UVM_IS_OK  and  st != UVM_HAS_X):
#            uvm_error("RegModel",
#                       sv.sformatf("Unable to read MSB bits in %s[%0d] to for RMW cycle on virtual field %s.",
#                                 mem.get_full_name(), segoff+segn-1,
#                                 self.get_full_name()))
#            status = UVM_NOT_OK
#            self.parent.XatomicX(0)
#            return
#         end
#      end
#      value |= (tmp & ~((1<<rmwbits)-1)) << ((segn-1)*segsiz)
#   end
#
#   // Now write each of the segments
#   tmp = value
#   repeat (segn):
#      mem.write(st, segoff, tmp, path, map, parent,, extension, fname, lineno)
#      if (st != UVM_IS_OK  and  st != UVM_HAS_X) status = UVM_NOT_OK
#
#      segoff++
#      tmp = tmp >> segsiz
#   end
#
#   self.post_write(idx, value, path, map, status)
#   for (uvm_vreg_field_cbs cb = cbs.first(); cb is not None
#        cb = cbs.next()):
#      cb.fname = self.fname
#      cb.lineno = self.lineno
#      cb.post_write(self, idx, value, path, map, status)
#   end
#
#   self.parent.XatomicX(0)
#
#
#   uvm_info("RegModel", sv.sformatf("Wrote virtual field \"%s\"[%0d] via %s with: 'h%h",
#                              self.get_full_name(), idx,
#                              (path == UVM_FRONTDOOR) ? "frontdoor" : "backdoor",
#                              value),UVM_MEDIUM);
#
#   write_in_progress = 1'b0
#   self.fname = ""
#   self.lineno = 0
#endtask: write
#
#
#@cocotb.coroutine
#task UVMVRegField::read(input longint unsigned     idx,
#                          output uvm_status_e   status,
#                          output uvm_reg_data_t      value,
#                          input  uvm_path_e     path = UVM_DEFAULT_PATH,
#                          input  uvm_reg_map      map = None,
#                          input  uvm_sequence_base   parent = None,
#                          input  uvm_object          extension = None,
#                          input  string              fname = "",
#                          input  int                 lineno = 0)
#   uvm_reg_data_t  tmp
#   uvm_reg_data_t  segval
#   uvm_reg_addr_t  segoff
#   uvm_status_e st
#
#   int flsb, lsb
#   int segsiz, segn
#   uvm_mem    mem
#
#   uvm_vreg_field_cb_iter cbs = new(self)
#
#   self.fname = fname
#   self.lineno = lineno
#
#   read_in_progress = 1'b1
#   mem = self.parent.get_memory()
#   if (mem is None):
#      uvm_error("RegModel", sv.sformatf("Cannot call UVMVRegField::read() on unimplemented virtual register \"%s\"",
#                                     self.get_full_name()))
#      status = UVM_NOT_OK
#      return
#   end
#
#   if (path == UVM_DEFAULT_PATH):
#      uvm_reg_block blk = self.parent.get_block()
#      path = blk.get_default_path()
#   end
#
#   status = UVM_IS_OK
#
#   self.parent.XatomicX(1)
#
#   value = 0
#
#   self.pre_read(idx, path, map)
#   for (uvm_vreg_field_cbs cb = cbs.first(); cb is not None
#        cb = cbs.next()):
#      cb.fname = self.fname
#      cb.lineno = self.lineno
#      cb.pre_read(self, idx, path, map)
#   end
#
#   segsiz = mem.get_n_bytes() * 8
#   flsb    = self.get_lsb_pos_in_register()
#   segoff  = self.parent.get_offset_in_memory(idx) + (flsb / segsiz)
#   lsb = flsb % segsiz
#
#   // Total number of memory segment in this field
#   segn = (lsb + self.get_n_bits() - 1) / segsiz + 1
#
#   // Read each of the segments, MSB first
#   segoff += segn - 1
#   repeat (segn):
#      value = value << segsiz
#
#      mem.read(st, segoff, tmp, path, map, parent, , extension, fname, lineno)
#      if (st != UVM_IS_OK  and  st != UVM_HAS_X) status = UVM_NOT_OK
#
#      segoff--
#      value |= tmp
#   end
#
#   // Any bits on the LSB side we need to get rid of?
#   value = value >> lsb
#
#   // Any bits on the MSB side we need to get rid of?
#   value &= (1<<self.get_n_bits()) - 1
#
#   self.post_read(idx, value, path, map, status)
#   for (uvm_vreg_field_cbs cb = cbs.first(); cb is not None
#        cb = cbs.next()):
#      cb.fname = self.fname
#      cb.lineno = self.lineno
#      cb.post_read(self, idx, value, path, map, status)
#   end
#
#   self.parent.XatomicX(0)
#
#   uvm_info("RegModel", sv.sformatf("Read virtual field \"%s\"[%0d] via %s: 'h%h",
#                              self.get_full_name(), idx,
#                              (path == UVM_FRONTDOOR) ? "frontdoor" : "backdoor",
#                              value),UVM_MEDIUM)
#
#
#   read_in_progress = 1'b0
#   self.fname = ""
#   self.lineno = 0
#endtask: read
#
#
#@cocotb.coroutine
#task UVMVRegField::poke(input  longint unsigned  idx,
#                          output uvm_status_e status,
#                          input  uvm_reg_data_t    value,
#                          input  uvm_sequence_base parent = None,
#                          input  uvm_object        extension = None,
#                          input  string            fname = "",
#                          input  int               lineno = 0)
#   uvm_reg_data_t  tmp
#   uvm_reg_data_t  segval
#   uvm_reg_addr_t  segoff
#   uvm_status_e st
#
#   int flsb, fmsb, rmwbits
#   int segsiz, segn
#   uvm_mem    mem
#   uvm_path_e rm_path
#   self.fname = fname
#   self.lineno = lineno
#
#   mem = self.parent.get_memory()
#   if (mem is None):
#      uvm_error("RegModel", sv.sformatf("Cannot call UVMVRegField::poke() on unimplemented virtual register \"%s\"",
#                                     self.get_full_name()))
#      status = UVM_NOT_OK
#      return
#   end
#
#   status = UVM_IS_OK
#
#   self.parent.XatomicX(1)
#
#   if (value >> self.size):
#      uvm_warning("RegModel", sv.sformatf("Writing value 'h%h that is greater than field \"%s\" size (%0d bits)", value, self.get_full_name(), self.get_n_bits()))
#      value &= value & ((1<<self.size)-1)
#   end
#   tmp = 0
#
#   segsiz = mem.get_n_bytes() * 8
#   flsb    = self.get_lsb_pos_in_register()
#   segoff  = self.parent.get_offset_in_memory(idx) + (flsb / segsiz)
#
#   // Any bits on the LSB side we need to RMW?
#   rmwbits = flsb % segsiz
#
#   // Total number of memory segment in this field
#   segn = (rmwbits + self.get_n_bits() - 1) / segsiz + 1
#
#   if (rmwbits > 0):
#      uvm_reg_addr_t  segn
#
#      mem.peek(st, segoff, tmp, "", parent, extension, fname, lineno)
#      if (st != UVM_IS_OK  and  st != UVM_HAS_X):
#         uvm_error("RegModel",
#                    sv.sformatf("Unable to read LSB bits in %s[%0d] to for RMW cycle on virtual field %s.",
#                              mem.get_full_name(), segoff, self.get_full_name()))
#         status = UVM_NOT_OK
#         self.parent.XatomicX(0)
#         return
#      end
#
#      value = (value << rmwbits) | (tmp & ((1<<rmwbits)-1))
#   end
#
#   // Any bits on the MSB side we need to RMW?
#   fmsb = rmwbits + self.get_n_bits() - 1
#   rmwbits = (fmsb+1) % segsiz
#   if (rmwbits > 0):
#      if (segn > 0):
#         mem.peek(st, segoff + segn - 1, tmp, "", parent, extension, fname, lineno)
#         if (st != UVM_IS_OK  and  st != UVM_HAS_X):
#            uvm_error("RegModel",
#                       sv.sformatf("Unable to read MSB bits in %s[%0d] to for RMW cycle on virtual field %s.",
#                                 mem.get_full_name(), segoff+segn-1,
#                                 self.get_full_name()))
#            status = UVM_NOT_OK
#            self.parent.XatomicX(0)
#            return
#         end
#      end
#      value |= (tmp & ~((1<<rmwbits)-1)) << ((segn-1)*segsiz)
#   end
#
#   // Now write each of the segments
#   tmp = value
#   repeat (segn):
#      mem.poke(st, segoff, tmp, "", parent, extension, fname, lineno)
#      if (st != UVM_IS_OK  and  st != UVM_HAS_X) status = UVM_NOT_OK
#
#      segoff++
#      tmp = tmp >> segsiz
#   end
#
#   self.parent.XatomicX(0)
#
#   uvm_info("RegModel", sv.sformatf("Wrote virtual field \"%s\"[%0d] with: 'h%h",
#                              self.get_full_name(), idx, value),UVM_MEDIUM)
#
#   self.fname = ""
#   self.lineno = 0
#endtask: poke
#
#
#@cocotb.coroutine
#task UVMVRegField::peek(input  longint unsigned  idx,
#                          output uvm_status_e status,
#                          output uvm_reg_data_t    value,
#                          input  uvm_sequence_base parent = None,
#                          input  uvm_object        extension = None,
#                          input  string            fname = "",
#                          input  int               lineno = 0)
#   uvm_reg_data_t  tmp
#   uvm_reg_data_t  segval
#   uvm_reg_addr_t  segoff
#   uvm_status_e st
#
#   int flsb, lsb
#   int segsiz, segn
#   uvm_mem    mem
#   self.fname = fname
#   self.lineno = lineno
#
#   mem = self.parent.get_memory()
#   if (mem is None):
#      uvm_error("RegModel", sv.sformatf("Cannot call UVMVRegField::peek() on unimplemented virtual register \"%s\"",
#                                     self.get_full_name()))
#      status = UVM_NOT_OK
#      return
#   end
#
#   status = UVM_IS_OK
#
#   self.parent.XatomicX(1)
#
#   value = 0
#
#   segsiz = mem.get_n_bytes() * 8
#   flsb    = self.get_lsb_pos_in_register()
#   segoff  = self.parent.get_offset_in_memory(idx) + (flsb / segsiz)
#   lsb = flsb % segsiz
#
#   // Total number of memory segment in this field
#   segn = (lsb + self.get_n_bits() - 1) / segsiz + 1
#
#   // Read each of the segments, MSB first
#   segoff += segn - 1
#   repeat (segn):
#      value = value << segsiz
#
#      mem.peek(st, segoff, tmp, "", parent, extension, fname, lineno)
#
#      if (st != UVM_IS_OK  and  st != UVM_HAS_X) status = UVM_NOT_OK
#
#      segoff--
#      value |= tmp
#   end
#
#   // Any bits on the LSB side we need to get rid of?
#   value = value >> lsb
#
#   // Any bits on the MSB side we need to get rid of?
#   value &= (1<<self.get_n_bits()) - 1
#
#   self.parent.XatomicX(0)
#
#   uvm_info("RegModel", sv.sformatf("Peeked virtual field \"%s\"[%0d]: 'h%h", self.get_full_name(), idx, value),UVM_MEDIUM)
#
#   self.fname = ""
#   self.lineno = 0
#endtask: peek
#
#
#def void UVMVRegField::do_print (self,uvm_printer printer):
#  super().do_print(printer)
#  printer.print_generic("initiator", parent.get_type_name(), -1, convert2string())
#endfunction
#
#def string UVMVRegField::convert2string(self):
#   string res_str
#   string t_str
#   bit with_debug_info = 1'b0
#   $sformat(convert2string, {"%s[%0d-%0d]"},
#            self.get_name(),
#            self.get_lsb_pos_in_register() + self.get_n_bits() - 1,
#            self.get_lsb_pos_in_register())
#   if (read_in_progress == 1'b1):
#      if (fname != ""  and  lineno != 0)
#         $sformat(res_str, "%s:%0d ",fname, lineno)
#      convert2string = {convert2string, "\n", res_str, "currently executing read method"};
#   end
#   if ( write_in_progress == 1'b1):
#      if (fname != ""  and  lineno != 0)
#         $sformat(res_str, "%s:%0d ",fname, lineno)
#      convert2string = {convert2string, "\n", res_str, "currently executing write method"};
#   end
#
#endfunction
#
#//TODO - add fatal messages
#
#def uvm_object UVMVRegField::clone(self):
#  return None
#endfunction
#
#def void UVMVRegField::do_copy   (self,uvm_object rhs):
#endfunction
#
#function bit UVMVRegField::do_compare (uvm_object  rhs,
#                                        uvm_comparer comparer)
#  return 0
#endfunction
#
#def void UVMVRegField::do_pack (self,uvm_packer packer):
#endfunction
#
#def void UVMVRegField::do_unpack (self,uvm_packer packer):
#endfunction



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc. 
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2014      NVIDIA Corporation
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------
#
#`ifndef UVM_VERSION_DEFINES_SVH
#`define UVM_VERSION_DEFINES_SVH
#
#// Title: UVM Version Defines
#
#// Group: UVM Revision Values
#// These macros provide the current values for the ~MAJOR~, ~MINOR~,
#// and optionally the ~FIX~ revision.
#//
#// Example with UVM version 1.2:
#//  UVM_MAJOR_REV - '1'
#//  UVM_MINOR_REV - '2'
#//  UVM_FIX_REV - 'undefined'
#//
#// Example with UVM version 1.1a:
#//  UVM_MAJOR_REV - '1'
#//  UVM_MINOR_REV - '1'
#//  UVM_FIX_REV - 'a'
#//
#
#// Macro: UVM_MAJOR_REV
#// Defines the MAJOR revision number.
#//
#// For UVM version 1.2, the MAJOR revision number is '1'
#//
#// | `define UVM_MAJOR_REV 1
#`define UVM_MAJOR_REV 1
#
#// Macro: UVM_MINOR_REV
#// Defines the MINOR revision number.
#//
#// For UVM version 1.2, the MINOR revision number is '2'
#//
#// | `define UVM_MINOR_REV 2
#`define UVM_MINOR_REV 2
#
#// Macro: UVM_FIX_REV
#// (Optionally) Defines the FIX revision letter.
#//
#// For the first "X.Y" release of the UVM, there is no
#// FIX revision letter.  In these cases, the UVM_FIX_REV
#// is left undefined.
#//
#// For any subsequent "X.Y" fix releases, the UVM_FIX_REV
#// value is set to the appropriate fix release letter.
#//
#// Example:
#// 1.1 - First release, UVM_FIX_REV is undefined
#// 1.1a - Fix release, UVM_FIX_REV is 'a'
#
#// Include this in the comment when the FIX rev is present:
#// | `define UVM_FIX_REV a
#
#//`define UVM_FIX_REV a
#
#// Macro: UVM_NAME
#// The name used by the library when displaying the name of
#// the library.
#//
#// | `define UVM_NAME UVM
#`define UVM_NAME UVM
#
#// Macro: UVM_VERSION_STRING
#// Provides a string-ized version of the UVM Library version number.
#//
#// When there is a FIX_REV, the string is "<name>-<major>.<minor><fix>" (such as "UVM-1.1d").
#// When there is NO FIX_REV, the string is "<name>-<major>.<minor>" (such as "UVM-1.2").
#
#`ifdef UVM_FIX_REV
# `define UVM_VERSION_STRING `"`UVM_NAME``-```UVM_MAJOR_REV``.```UVM_MINOR_REV`UVM_FIX_REV`"
#`else
# `define UVM_VERSION_STRING `"`UVM_NAME``-```UVM_MAJOR_REV``.```UVM_MINOR_REV```"
#`endif
#
#// Group: Conditional Compilation
#// These macros provide the ability to conditionally compile based on the
#// revision of the library which is being used.
#//
#// These macros are required for conditional compilation, as SystemVerilog does
#// not support conditionals with `ifdefs.
#//
#// For example:
#// |  // Illegal:
#// | `if (UVM_MAJOR_REV == 1)
#// |
#// | // Legal:
#// | `ifdef UVM_MAJOR_REV_1
#// 
#
#
#// Macro: UVM_MAJOR_REV_1
#// Indicates that the MAJOR version of this release is '1'.
#//
#// | `define UVM_MAJOR_REV_1
#//
#`define UVM_MAJOR_REV_1
#
#// Macro: UVM_MINOR_REV_2
#// Indicates that the MINOR version of this release is '2'.
#//
#// | `define UVM_MINOR_REV_2
#//
#`define UVM_MINOR_REV_2
#
#// Macro: UVM_VERSION_1_2
#// Indicates that the version of this release is '1.2'.
#//
#// | `define UVM_VERSION_1_2
#`define UVM_VERSION_1_2
#
#// Undocumented, same thing as UVM_VERSION_1_2
#`define UVM_MAJOR_VERSION_1_2
#
#// Group: UVM Version Ladder
#
#// Macro: UVM_POST_VERSION_1_1
#// Indicates that this version of the UVM came after the
#// 1.1 versions, including the various 1.1 fix revisions.  
#//
#// The first UVM version wherein this macro is defined is
#// 1.2, and the macro will continue to be defined
#// for all future revisions of the UVM library.
#//
#// | `define UVM_POST_VERSION_1_1
#`define UVM_POST_VERSION_1_1
#
#`endif // UVM_VERSION_DEFINES_SVH



#
#//----------------------------------------------------------------------
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

# About: master_slave_test
#
# This toy example uses TLM master/slave ports and transport-interface.
#
# Walk through the test:
# ======================
# 1. Master comps will generate NUM_ITEMS transactions.
#
# 2. Bus will accept the transactions, and route them to correct slaves or
#    memories.
#
# 3. Memories respond to writes/reads in blocking manner. Slave also does the
#    same. IO Slave uses transport to communicate with an IO port.
#
# 4. IO Port blocks on READ but WRITES are non-blocking (shoot and forget)>
#
# 5. Top_env does some checks at the end to ensure some transactions have been
#    completed.

import cocotb
from cocotb.triggers import Timer

from uvm import *
from enum import Enum

# These can be changed to vary the topology
NUM_MASTERS = 5  # 6 fatals out
NUM_SLAVES = 34  # 35 gets stuck, >= 36 fatals out
NUM_MEMS = 10 # 11 gets stuck

#ratio = 3
#NUM_MASTERS = 2 * ratio
#NUM_SLAVES = 4 * ratio
#NUM_MEMS = 4 * ratio


class Cmd(Enum):
    READ = 0
    WRITE = 1


NUM_ITEMS = 100
MEM_ADDR_MASK = 0xFFFFF
MEM_ADDR_OFFSET = MEM_ADDR_MASK + 1
SLAVE_ADDR_OFFSET = 0x1000


def get_slv_addr(i):
    addr = SLAVE_ADDR_OFFSET + i * SLAVE_ADDR_OFFSET
    if addr >= MEM_ADDR_OFFSET:
        raise Exception("slave_addr {} out of range (<= {}), i: {}".format(addr,
            MEM_ADDR_OFFSET, i))
    return addr


def get_master_addr(i):
    addr = 0x10 + i * 0x10
    if addr >= SLAVE_ADDR_OFFSET:
        raise Exception("master_addr {} out of range (<= {})".format(addr,
            SLAVE_ADDR_OFFSET))
    return addr



def get_mem_addr(i):
    return MEM_ADDR_OFFSET + i * MEM_ADDR_OFFSET


#----------------------------------------------------------------------
# class transaction
#----------------------------------------------------------------------


class transaction(UVMTransaction):


    def __init__(self, name='trans'):
        super().__init__(name)
        self.addr = 0
        self.data = 0
        self.cmd = Cmd.READ
        addr_list = []
        for i in range(NUM_SLAVES):
            addr_list.append(get_slv_addr(i))
        for i in range(NUM_MEMS):
            addr_list.append(get_mem_addr(i))
        self.rand("data", range((1 << 32) - 1))
        self.rand("addr", addr_list)
        self.rand("cmd", [Cmd.READ, Cmd.WRITE])

    def do_copy(self, t):
        self.data = t.data
        self.addr = t.addr

    def do_compare(self, a, b):
        return a.data == b.data and a.addr == b.addr

    def do_clone(self):
        t = transaction()
        t.copy(self)
        return t

    def convert2string(self):
        s = sv.sformatf("[ addr = %s, data = %s cmd: %s]", hex(self.addr),
                hex(self.data), str(self.cmd))
        return s


uvm_object_utils(transaction)


#----------------------------------------------------------------------
# component master
#----------------------------------------------------------------------

class master(UVMComponent):

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.master_port = UVMMasterPort("master_port", self)
        self.num_items = 0
        self.use_blocking = True
        self.tag = "MASTER"

    
    async def run_phase(self, phase):
        for i in range(NUM_ITEMS):
            t = transaction()
            if not t.randomize():
                uvm_error("GEN", "Failed to randomize transaction")
            else:
                t.addr = t.addr
            uvm_info(self.tag, sv.sformatf("sending  : %s", t.convert2string()), UVM_MEDIUM)
            if t.cmd == Cmd.WRITE:
                await self.master_port.put(t)
            else:
                await self.master_port.get([t])
            self.num_items += 1
            await Timer(sv.urandom_range(1, 10) * 10, "NS")


# Base class for slave components

class slave_base(UVMComponent):

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.slave_port = UVMSlaveImp('slave_port', self)

    def try_put(self, t):
        return False

    def can_put(self, t):
        return False

    def try_get(self, t):
        return False

    def can_get(self, t):
        return False

    def try_peek(self, t):
        return False

    def can_peek(self, t):
        return False

    
    async def get(self, t):
        await Timer(0)

    
    async def put(self, t):
        await Timer(0)

    
    async def peek(self, t):
        await Timer(0)


class bus(slave_base):
    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.addr_map = {}
        self.req_fifo = UVMTLMFIFO('bus_fifo')
        self.master_ports = {}
        self.master_port = UVMMasterImp('master_port', self)

    def add_comp(self, addr, comp, port_name=''):
        port = None
        # if addr in self.addr_map:
        if port_name == "":
            if hasattr(comp, 'master_port'):
                port_name = 'master_port'
                port = getattr(comp, 'master_port')
            elif hasattr(comp, 'slave_port'):
                port_name = 'slave_port'
                port = getattr(comp, 'slave_port')
            else:
                uvm_fatal("BUS/NO_PORT",
                    "Unable to find master/slave port, and no port name given!")
        # pname = port  # TODO support for any port name
        if port is not None:
            self.addr_map[addr] = {"comp": comp}
            if port_name == 'master_port':
                port.connect(self.master_port)
            else:
                self.master_ports[addr] = UVMSlavePort('bus_port_' +
                        comp.get_name(), self)
                self.master_ports[addr].connect(port)

    
    async def get(self, t):
        addr = t[0].addr
        if addr in self.addr_map:
            await self.master_ports[addr].get(t)
        else:
            uvm_error('BUS_READ_ERR', "No addr {} in map".format(addr))

    
    async def put(self, t):
        addr = t.addr
        if addr in self.addr_map:
            await self.master_ports[addr].put(t)
        else:
            uvm_error('BUS_WRITE_ERR', "No addr {} in map".format(addr))

#----------------------------------------------------------------------
# componenet slave
#----------------------------------------------------------------------


class slave(slave_base):


    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.num_put = 0
        self.num_get = 0

    def num_cmds(self):
        return self.num_put + self.num_get

    
    async def put(self, t):
        await Timer(10, "NS")
        uvm_info("SLAVE_PUT", "Received cmd", UVM_MEDIUM)
        self.num_put += 1


    
    async def get(self, t):
        await Timer(10, "NS")
        uvm_info("SLAVE_GET", "Received cmd", UVM_MEDIUM)
        self.num_get += 1


class io_slave(slave):

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.trans_port = UVMTransportPort('trans_imp', self)

    
    async def put(self, t):
        await Timer(10, "NS")
        uvm_info("SLAVE_PUT", "Received cmd", UVM_MEDIUM)
        self.num_put += 1
        rsp = []
        if self.trans_port.nb_transport(t, rsp):
            uvm_info(self.tag, "Transport put to IO port OK")


    
    async def get(self, t):
        await Timer(10, "NS")
        uvm_info("IO_SLAVE_GET",
            "Oh no! We hit the blocking IO slave read. This will take long time",
            UVM_MEDIUM)
        self.num_get += 1
        rsp = []
        await self.trans_port.transport(t[0], rsp)
        t.clear()
        t.append(rsp[0])


class io_port(UVMComponent):

    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.trans_imp = UVMTransportImp('trans_imp', self)
        self.num_io_writes = 0
        self.num_io_reads = 0

    # IO Reads implemented as blocking
    
    async def transport(self, req, rsp):
        await Timer(500, "NS")
        self.num_io_reads += 1
        rsp.append("READ DONE")

    # Writes implemented as non-blocking
    def nb_transport(self, req, rsp):
        self.num_io_writes += 1
        # Data goes to /dev/null etc
        rsp.append("WRITE DONE")

#  //----------------------------------------------------------------------
#  // componenet bfm
#  //----------------------------------------------------------------------

class memory(slave_base):


    def __init__(self, name, parent, size=256):
        super().__init__(name, parent)
        self.size = 256
        self.mem = {}
        self.num_reads = 0
        self.num_writes = 0
        for i in range(self.size):
            self.mem[i] = 0x0

    def num_cmds(self):
        return self.num_reads + self.num_writes

    
    async def put(self, t):
        await self.wait_access()
        addr = t.addr & MEM_ADDR_MASK
        if addr in self.mem:
            self.num_writes += 1
            uvm_info("MEM_WRITE", "Wrote data [{:0X}]: {:0X}".format(addr,
                t.data), UVM_MEDIUM)
            self.mem[addr] = t.data

    
    async def get(self, t):
        await self.wait_access()
        addr = t[0].addr & MEM_ADDR_MASK
        if addr in self.mem:
            self.num_reads += 1
            t.append(self.mem[addr])
            uvm_info("MEM_READ", "Read data [{:0X}]: {:0X}".format(addr,
                self.mem[addr]), UVM_MEDIUM)

    
    async def wait_access(self):
        await Timer(10, "NS")


    
    async def run_phase(self, phase):
        while True:
            await Timer(100, "NS")


class listener(UVMSubscriber):  # (transaction)
    def __init__(self, name, parent):
        super().__init__(name,parent)

    def write(self, t):
        uvm_info(m_name, sv.sformatf("Received: %s", t.convert2string()), UVM_MEDIUM)


#  //----------------------------------------------------------------------
#  // componenet top
#  //----------------------------------------------------------------------

class env_top(UVMEnv):


    def __init__(self, name, parent):
        super().__init__(name, parent)
        self.masters = []
        self.slaves = []
        self.mems = []
        self.bus = None

    def build_phase(self, phase):
        self.bus = bus("bus", self)
        # Need to call these in build() because this creates some TLM ports
        for i in range(NUM_MASTERS):
            mast = master("master_" + str(i), self)
            mast.tag = mast.tag + "__" + str(i)
            self.masters.append(mast)
            addr = get_master_addr(i)
            self.bus.add_comp(addr, mast)

        for i in range(NUM_SLAVES):
            slv = None
            if sv.urandom_range(0, 9) == 0:
                slv = io_slave("io_slave_" + str(i), self)
                io_port_inst = io_port("io_port_" + str(i), self)
                # P2P connection, not done via bus
                slv.trans_port.connect(io_port_inst.trans_imp)
            else:
                slv = slave("slave_" + str(i), self)
            self.slaves.append(slv)
            addr = get_slv_addr(i)
            self.bus.add_comp(addr, slv)
        for i in range(NUM_MEMS):
            mem_inst = memory("mem_" + str(i), self)
            self.mems.append(mem_inst)
            self.bus.add_comp(get_mem_addr(i), mem_inst)


    def connect_phase(self, phase):
        pass

    
    async def run_phase(self, phase):
        phase.raise_objection(self)
        uvm_info("ENV_TOP", "run_phase started", UVM_MEDIUM)
        await Timer(NUM_ITEMS * 110, "NS")
        phase.drop_objection(self)


    def all_ok(self):
        num_master_cmds = NUM_MASTERS * NUM_ITEMS
        num_slave_cmds = 0
        num_mem_cmds = 0
        for i in range(NUM_SLAVES):
            num_slave_cmds += self.slaves[i].num_cmds()
        for i in range(NUM_MEMS):
            num_mem_cmds += self.mems[i].num_cmds()

        if (num_slave_cmds + num_mem_cmds) != num_master_cmds:
            slv_cmd = num_slave_cmds + num_mem_cmds
            uvm_error("ENV_TOP", "Cmd mismatch: {} vs master: {}".format(
                slv_cmd, num_master_cmds))
            return False
        else:
            return True
# uvm_component_utils(env_top)



#
#------------------------------------------------------------------------------
#   Copyright 2007-2010 Mentor Graphics Corporation
#   Copyright 2007-2010 Cadence Design Systems, Inc.
#   Copyright 2010 Synopsys, Inc.
#   Copyright 2019 Tuomas Poikela
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http:#www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
# Title: Factory Component and Object Wrappers
#
# Topic: Intro
#
# This section defines the proxy component and object classes used by the
# factory. To avoid the overhead of creating an instance of every component
# and object that get registered, the factory holds lightweight wrappers,
# or proxies. When a request for a new object is made, the factory calls upon
# the proxy to create the object it represents.
#------------------------------------------------------------------------------

#------------------------------------------------------------------------------
#
# CLASS: uvm_component_registry #(T,Tname)
#
# The uvm_component_registry serves as a lightweight proxy for a component of
# type ~T~ and type name ~Tname~, a string. The proxy enables efficient
# registration with the <uvm_factory>. Without it, registration would
# require an instance of the component itself.
#
# See <Usage> section below for information on using uvm_component_registry.
#
#------------------------------------------------------------------------------

from .uvm_factory import UVMObjectWrapper
from .uvm_globals import uvm_report_warning, uvm_report_fatal
from .uvm_object_globals import UVM_NONE


def get_factory():
    from .uvm_coreservice import UVMCoreService
    cs = UVMCoreService.get()
    return cs.get_factory()


class UVMComponentRegistry(UVMObjectWrapper):

    # dict of UVMComponentRegistry
    registry_db = {}  # tname -> UVMComponentRegistry
    registered = {}  # tname -> bool
    # Stores the actual constructors
    comps = {}

    def __init__(self, Constr, tname):
        self.Constr = Constr
        self.tname = tname
        self.type_name = tname
        if tname in UVMComponentRegistry.registry_db:
            uvm_report_warning("COMP_REGD", (tname + ' has already been'
                + ' added into the UVMComponentRegistry'))
        UVMComponentRegistry.comps[tname] = Constr
        UVMComponentRegistry.registered[tname] = False
        UVMComponentRegistry.registry_db[tname] = self.get()

    @classmethod
    def reset(cls):
        """Resets the state of the component registry. Used for unit testing"""
        UVMComponentRegistry.registry_db.clear()
        UVMComponentRegistry.registered.clear()
        UVMComponentRegistry.comps.clear()

    def create_component(self, name, parent):
        """
        Function: create_component

        Creates a component of type T having the provided `name` and `parent`.
        This is an override of the method in `uvm_object_wrapper`. It is
        called by the factory after determining the type of object to create.
        You should not call this method directly. Call `create` instead.
        Args:
            name:
            parent:
        Returns:
        """
        return UVMComponentRegistry.comps[self.tname](name, parent)

    def get_type_name(self):
        """
        Function: get_type_name

        Returns the value given by the string parameter, `Tname`. This method
        overrides the method in `uvm_object_wrapper`.
        Returns:
        """
        return self.tname

    def get(self):
        """
        Function: get

        Returns the singleton instance of this type. Type-based factory operation
        depends on there being a single proxy instance for each registered type.
        Returns:
        """
        if UVMComponentRegistry.registered[self.tname] is False:
            factory = get_factory()
            factory.register(self)
            UVMComponentRegistry.registered[self.tname] = True
        return self

    def create(self, name, parent, contxt=""):
        """
        Function: create

        Returns an instance of the component type, `T`, represented by this proxy,
        subject to any factory overrides based on the context provided by the
        `parent`'s full name. The `contxt` argument, if supplied, supersedes the
        `parent`'s context. The new instance will have the given leaf `name`
        and `parent`.
        Args:
            name:
            parent:
            contxt:
        Returns:
        Raises:
        """
        factory = get_factory()

        if contxt == "" and parent is not None:
            contxt = parent.get_full_name()
        obj = factory.create_component_by_type(self.get(),contxt,name,parent)
        if obj is not None:
            return obj

        # cast check removed, TSP
        msg = ("Factory did not return a component of type '" + self.type_name
          + "'. A component of type '")
        if obj is None:
            msg += "null"
        else:
            get_type_name_func = getattr(obj, "get_type_name", None)
            if callable(get_type_name_func):
                msg += obj.get_type_name()
            else:
                klass_name = obj.__class__
                raise Exception('No get_type_name in {}. Did you use uvm_componen_utils({})'.format(
                    klass_name, klass_name))

        msg += "' was returned instead. Name=" + name + " Parent="
        if parent is None:
            msg += 'null'
        else:
            msg += parent.get_type_name()
        msg += " contxt=" + contxt
        uvm_report_fatal("FCTTYP", msg, UVM_NONE)

    def set_type_override(self, override_type, replace=True):
        """
        Function: set_type_override

        Configures the factory to create an object of the type represented by
        `override_type` whenever a request is made to create an object of the type,
        `T`, represented by this proxy, provided no instance override applies. The
        original type, `T`, is typically a super class of the override type.
        Args:
            override_type:
            replace:
        """
        factory = get_factory()
        factory.set_type_override_by_type(self.get(),override_type,replace)

    def set_inst_override(self, override_type, inst_path, parent=None):
        """
        Function: set_inst_override

        Configures the factory to create a component of the type represented by
        `override_type` whenever a request is made to create an object of the type,
        `T`, represented by this proxy,  with matching instance paths. The original
        type, `T`, is typically a super class of the override type.

        If `parent` is not specified, `inst_path` is interpreted as an absolute
        instance path, which enables instance overrides to be set from outside
        component classes. If `parent` is specified, `inst_path` is interpreted
        as being relative to the `parent`'s hierarchical instance path, i.e.
        ~{parent.get_full_name(),".",inst_path}~ is the instance path that is
        registered with the override. The `inst_path` may contain wildcards for
        matching against multiple contexts.
        Args:
            override_type:
            inst_path:
            parent:
        """
        if parent is not None:
            if inst_path == "":
                inst_path = parent.get_full_name()
            else:
                inst_path = parent.get_full_name() + "." + inst_path

        factory = get_factory()
        factory.set_inst_override_by_type(self.get(),override_type,inst_path)


class UVMObjectRegistry(UVMObjectWrapper):

    # dict of UVMObjectRegistry
    registry_db = {}  # tname -> UVMObjectRegistry
    registered = {}  # tname -> bool
    # Stores the actual constructors
    objs = {}

    def __init__(self, Constr, tname):
        self.Constr = Constr
        self.tname = tname
        self.type_name = tname
        if tname in UVMObjectRegistry.registry_db:
            uvm_report_warning("COMP_REGD", (tname + ' has already been'
                + ' added into the UVMObjectRegistry'))
        UVMObjectRegistry.objs[tname] = Constr
        UVMObjectRegistry.registered[tname] = False
        UVMObjectRegistry.registry_db[tname] = self.get()

    @staticmethod
    def reset():
        """Resets the state of the object registry. Used for unit testing"""
        UVMObjectRegistry.registry_db.clear()
        UVMObjectRegistry.registered.clear()
        UVMObjectRegistry.objs.clear()

    # Function: create_object
    #
    # Creates an object of type ~T~ and returns it as a handle to a
    # <uvm_object>. This is an override of the method in <uvm_object_wrapper>.
    # It is called by the factory after determining the type of object to create.
    # You should not call this method directly. Call <create> instead.

    def create_object(self, name=""):
        return UVMObjectRegistry.objs[self.tname](name)

    def get_type_name(self):
        """
        Function: get_type_name

        Returns the value given by the string parameter, `Tname`. This method
        overrides the method in `uvm_object_wrapper`.
        Returns:
        """
        return self.type_name

    def get(self):
        """
        Function: get

        Returns the singleton instance of this type. Type-based factory operation
        depends on there being a single proxy instance for each registered type.
        Returns:
        """
        if UVMObjectRegistry.registered[self.tname] is False:
            factory = get_factory()
            factory.register(self)
            UVMObjectRegistry.registered[self.tname] = True
        return self

    def create(self, name, parent=None, contxt=""):
        """
        Function: create

        Returns an instance of the object type, `T`, represented by this proxy,
        subject to any factory overrides based on the context provided by the
        `parent`'s full name. The `contxt` argument, if supplied, supersedes the
        `parent`'s context. The new instance will have the given leaf `name`,
        if provided.
        Args:
            name:
            parent:
            contxt:
        Returns:
        """
        factory = get_factory()

        if contxt == "" and parent is not None:
            contxt = parent.get_full_name()
        obj = factory.create_object_by_type(self.get(),contxt,name)
        if obj is not None:
            return obj
        create = obj
        # tpoikela: TODO could check the type for extra safety
        #if (!$cast(create, obj)) begin
        #    string msg
        #    msg = {"Factory did not return an object of type '",type_name,
        #      "'. A component of type '",obj == null ? "null" : obj.get_type_name(),
        #      "' was returned instead. Name=",name," Parent=",
        #      parent==null?"null":parent.get_type_name()," contxt=",contxt}
        #    uvm_report_fatal("FCTTYP", msg, UVM_NONE)
        return create

    def set_type_override(self, override_type, replace=True):
        """
        Function: set_type_override

        Configures the factory to create an object of the type represented by
        `override_type` whenever a request is made to create an object of the type
        represented by this proxy, provided no instance override applies. The
        original type, `T`, is typically a super class of the override type.
        Args:
            override_type:
            replace:
        """
        factory = get_factory()
        factory.set_type_override_by_type(self.get(), override_type, replace)

    def set_inst_override(self, override_type, inst_path, parent=None):
        """
        Function: set_inst_override

        Configures the factory to create an object of the type represented by
        `override_type` whenever a request is made to create an object of the type
        represented by this proxy, with matching instance paths. The original
        type, `T`, is typically a super class of the override type.

        If `parent` is not specified, `inst_path` is interpreted as an absolute
        instance path, which enables instance overrides to be set from outside
        component classes. If `parent` is specified, `inst_path` is interpreted
        as being relative to the `parent`'s hierarchical instance path, i.e.
        ~{parent.get_full_name(),".",inst_path}~ is the instance path that is
        registered with the override. The `inst_path` may contain wildcards for
        matching against multiple contexts.
        Args:
            override_type:
            inst_path:
            parent:
        """
        factory = get_factory()

        if parent is not None:
            if inst_path == "":
                inst_path = parent.get_full_name()
            else:
                inst_path = parent.get_full_name() + "." + inst_path
        factory.set_inst_override_by_type(self.get(), override_type, inst_path)






#// 
#// -------------------------------------------------------------
#//    Copyright 2004-2011 Synopsys, Inc.
#//    Copyright 2010 Mentor Graphics Corporation
#//    Copyright 2010 Cadence Design Systems, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#// 
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#// 
#//        http://www.apache.org/licenses/LICENSE-2.0
#// 
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#// 

import cocotb
from cocotb.triggers import *

from uvm import sv
from uvm.seq import UVMSequence
from uvm.reg.uvm_reg_sequence import *
from uvm.macros import *

from blk_seqlib import *


class dut_reset_seq(UVMSequence):
    #
    def __init__(self, name="dut_reset_seq"):
        super().__init__(name)
        self.vif = None

    #
    
    async def body(self):
        self.vif.rst <= 1
        for i in range(5):
            await FallingEdge(self.vif.clk)
        self.vif.rst <= 0

    #endclass

uvm_object_utils(dut_reset_seq)


class sys_R_test_seq(UVMRegSequence):

    def __init__(self, name="sys_R_test_seq"):
        super().__init__(name)
        self.model = None  # reg_sys_S


    
    async def body(self):
        if self.model is None:
            uvm_fatal("NO_REG","RegModel model handle is None")
            return
        for i in range(2):
            await self.start_blk_seq(i)


    
    async def start_blk_seq(self, i):
        # fork
        seq = blk_R_test_seq.type_id.create(sv.sformatf("blk_seq%0d",i),
                None, self.get_full_name())
        seq.model = self.model.B[i]
        await seq.start(None, self)
        #end join_none


    #endclass

uvm_object_utils(sys_R_test_seq)



#//----------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2013 Cisco Systems, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

from .uvm_sequence_base import UVMSequenceBase
from ..macros.uvm_message_defines import uvm_fatal


class UVMSequence(UVMSequenceBase):
    """
    The UVMSequence class provides the interfaces necessary in order to create
    streams of sequence items and/or other sequences.
    """

    def __init__(self, name="uvm_sequence"):
        """
        Variable: req

        The sequence contains a field of the request type called req.  The user
        can use this field, if desired, or create another field to use.  The
        default `do_print` will print this field.

        Variable: rsp

        The sequence contains a field of the response type called rsp.  The user
        can use this field, if desired, or create another field to use.   The
        default `do_print` will print this field.

        Creates and initializes a new sequence object.

        Args:
            name:
        """
        UVMSequenceBase.__init__(self, name)
        self.req = None
        self.rsp = None
        self.param_sequencer = None

    #  // Function: send_request
    #  //
    #  // This method will send the request item to the sequencer, which will forward
    #  // it to the driver.  If the rerandomize bit is set, the item will be
    #  // randomized before being sent to the driver. The send_request function may
    #  // only be called after <uvm_sequence_base::wait_for_grant> returns.
    #
    #  function void send_request(uvm_sequence_item request, bit rerandomize = 0)
    #    REQ m_request
    #
    #    if (m_sequencer == null) begin
    #      uvm_report_fatal("SSENDREQ", "Null m_sequencer reference", UVM_NONE)
    #    end
    #    if (!$cast(m_request, request)) begin
    #      uvm_report_fatal("SSENDREQ", "Failure to cast uvm_sequence_item to request", UVM_NONE)
    #    end
    #    m_sequencer.send_request(this, m_request, rerandomize)
    #  endfunction


    #  // Function: get_current_item
    #  //
    #  // Returns the request item currently being executed by the sequencer. If the
    #  // sequencer is not currently executing an item, this method will return ~null~.
    #  //
    #  // The sequencer is executing an item from the time that get_next_item or peek
    #  // is called until the time that get or item_done is called.
    #  //
    #  // Note that a driver that only calls get will never show a current item,
    #  // since the item is completed at the same time as it is requested.
    #
    #  function REQ get_current_item()
    #    if (!$cast(param_sequencer, m_sequencer))
    #      uvm_report_fatal("SGTCURR", "Failure to cast m_sequencer to the parameterized sequencer", UVM_NONE)
    #    return (param_sequencer.get_current_item())
    #  endfunction


    #  // Task: get_response
    #  //
    #  // By default, sequences must retrieve responses by calling get_response.
    #  // If no transaction_id is specified, this task will return the next response
    #  // sent to this sequence.  If no response is available in the response queue,
    #  // the method will block until a response is received.
    #  //
    #  // If a transaction_id is parameter is specified, the task will block until
    #  // a response with that transaction_id is received in the response queue.
    #  //
    #  // The default size of the response queue is 8.  The get_response method must
    #  // be called soon enough to avoid an overflow of the response queue to prevent
    #  // responses from being dropped.
    #  //
    #  // If a response is dropped in the response queue, an error will be reported
    #  // unless the error reporting is disabled via
    #  // set_response_queue_error_report_disabled.
    async def get_response(self, response, transaction_id=-1):
        """
        Args:
            response:
            transaction_id:
        """
        if response is None:
            uvm_fatal("RESP IS NONE", "response arg must be an empty list")
        rsp = []
        await self.get_base_response(rsp, transaction_id)
        response.append(rsp[0])


    def put_response(self, response_item):
        """
        Internal method.

        Args:
            response_item (UVMSequenceItem): Response item
        """
        #response
        #if (!$cast(response, response_item)) begin
        #    uvm_report_fatal("PUTRSP", "Failure to cast response in put_response", UVM_NONE)
        self.put_base_response(response_item)


    def do_print(self, printer):
        """
        Function- do_print

        Args:
            printer (UVMPrinter): Printer used for printing
        """
        super().do_print(printer)
        printer.print_object("req", self.req)
        printer.print_object("rsp", self.rsp)



#//
#// -------------------------------------------------------------
#//    Copyright 2004-2011 Synopsys, Inc.
#//    Copyright 2010 Mentor Graphics Corporation
#//    Copyright 2010 Cadence Design Systems, Inc.
#//    Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//    All Rights Reserved Worldwide
#//
#//    Licensed under the Apache License, Version 2.0 (the
#//    "License"); you may not use this file except in
#//    compliance with the License.  You may obtain a copy of
#//    the License at
#//
#//        http://www.apache.org/licenses/LICENSE-2.0
#//
#//    Unless required by applicable law or agreed to in
#//    writing, software distributed under the License is
#//    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//    CONDITIONS OF ANY KIND, either express or implied.  See
#//    the License for the specific language governing
#//    permissions and limitations under the License.
#// -------------------------------------------------------------
#//

import cocotb
from cocotb.triggers import *

from uvm.reg.uvm_reg_sequence import *
from uvm.reg.uvm_reg_model import *
from uvm.base.sv import sv
from uvm.macros import *
from uvm.seq.uvm_sequence import *
from reg_B import *


class dut_reset_seq(UVMSequence):

    def __init__(self, name="dut_reset_seq"):
        super().__init__(name)
        self.vif = None


    
    async def body(self):
        self.vif.rst <= 1
        for i in range(5):
            await FallingEdge(self.vif.clk)
        self.vif.rst <= 0


uvm_object_utils(dut_reset_seq)


class blk_R_test_seq(UVMRegSequence):


    def __init__(self, name="blk_R_test_seq"):
        super().__init__(name)
        self.model = None  # reg_block_B


    def check_status(self, status, msg):
        if status[0] != UVM_IS_OK:
            uvm_fatal(msg, "status not OK")

    
    async def body(self):
        status = 0
        n = 0

        # Initialize R with a random value then check against mirror
        data = sv.urandom()
        status = []
        await self.model.R.write(status, data, parent=self)
        self.check_status(status, "WRITE_FAIL")
        status = []
        await self.model.R.mirror(status, UVM_CHECK, parent=self)
        self.check_status(status, "MIRROR FAIL")

        # Perform a random number of INC operations
        n = (sv.urandom() % 7) + 3
        uvm_info("blk_R_test_seq", sv.sformatf("Incrementing R %0d times...", n), UVM_NONE)
        for i in range(n):
            status = []
            await self.model.CTL.write(status, reg_fld_B_CTL_CTL.INC, parent=self)
            self.check_status(status, "WRITE " + str(i) + " FAIL")
            data += 1
            self.model.R.predict(data)
        # Check the final value
        status = []
        await self.model.R.mirror(status, UVM_CHECK, parent=self)
        self.check_status(status, "MIRROR2 FAIL")

        # Perform a random number of DEC operations
        n = (sv.urandom() % 8) + 2
        uvm_info("blk_R_test_seq", sv.sformatf("Decrementing R %0d times...", n), UVM_NONE)
        for i in range(n):
            status = []
            await self.model.CTL.write(status, reg_fld_B_CTL_CTL.DEC, parent=self)
            self.check_status(status, "WRITE B." + str(i) + " FAIL (DEC OP)")
            data -= 1
            self.model.R.predict(data)

        # Check the final value
        status = []
        await self.model.R.mirror(status, UVM_CHECK, parent=self)
        self.check_status(status, "MIRROR3 FAIL")
        #
        # Reset the register and check
        status = []
        await self.model.CTL.write(status, reg_fld_B_CTL_CTL.CLR, parent=self)
        self.check_status(status, "WRITE4 FAIL")
        self.model.R.predict(0)
        status = []
        await self.model.R.mirror(status, UVM_CHECK, parent=self)
        self.check_status(status, "MIRROR4 FAIL")
        #   endtask
    #

uvm_object_utils(blk_R_test_seq)



#//----------------------------------------------------------------------
#//   Copyright 2007-2011 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2019 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

from typing import Dict

import cocotb
from cocotb.triggers import Event

from ..base.sv import sv
from ..base.uvm_queue import UVMQueue
from .uvm_sequence_item import UVMSequenceItem
from ..base.uvm_object_globals import *
from ..base.uvm_globals import uvm_zero_delay
from ..macros import uvm_fatal, uvm_warning, uvm_error
from ..base.uvm_pool import UVMPool
from ..dap.uvm_get_to_lock_dap import uvm_get_to_lock_dap
from ..base.uvm_recorder import UVMRecorder
from uvm.base.sv import wait

SEQ_ERR1_MSG = "neither the item's sequencer nor dedicated sequencer has been supplied to start item in "

SeqItemQueue = UVMQueue[UVMSequenceItem]

#//------------------------------------------------------------------------------
#//
#// CLASS: uvm_sequence_base
#//
#// The uvm_sequence_base class provides the interfaces needed to create streams
#// of sequence items and/or other sequences.
#//
#// A sequence is executed by calling its <start> method, either directly
#// or invocation of any of the `uvm_do_* macros.
#//
#// Executing sequences via <start>:
#//
#// A sequence's <start> method has a ~parent_sequence~ argument that controls
#// whether <pre_do>, <mid_do>, and <post_do> are called *in the parent*
#// sequence. It also has a ~call_pre_post~ argument that controls whether its
#// <pre_body> and <post_body> methods are called.
#// In all cases, its <pre_start> and <post_start> methods are always called.
#//
#// When <start> is called directly, you can provide the appropriate arguments
#// according to your application.
#//
#// The sequence execution flow looks like this
#//
#// User code
#//
#//| sub_seq.randomize(...); // optional
#//| sub_seq.start(seqr, parent_seq, priority, call_pre_post)
#//|
#//
#// The following methods are called, in order
#//
#//|
#//|   sub_seq.pre_start()        (task)
#//|   sub_seq.pre_body()         (task)  if call_pre_post==1
#//|     parent_seq.pre_do(0)     (task)  if parent_sequence!=None
#//|     parent_seq.mid_do(this)  (func)  if parent_sequence!=None
#//|   sub_seq.body               (task)  YOUR STIMULUS CODE
#//|     parent_seq.post_do(this) (func)  if parent_sequence!=None
#//|   sub_seq.post_body()        (task)  if call_pre_post==1
#//|   sub_seq.post_start()       (task)
#//
#//
#// Executing sub-sequences via `uvm_do macros:
#//
#// A sequence can also be indirectly started as a child in the <body> of a
#// parent sequence. The child sequence's <start> method is called indirectly
#// by invoking any of the `uvm_do macros.
#// In these cases, <start> is called with
#// ~call_pre_post~ set to 0, preventing the started sequence's <pre_body> and
#// <post_body> methods from being called. During execution of the
#// child sequence, the parent's <pre_do>, <mid_do>, and <post_do> methods
#// are called.
#//
#// The sub-sequence execution flow looks like
#//
#// User code
#//
#//|
#//| `uvm_do_with_prior(seq_seq, { constraints }, priority)
#//|
#//
#// The following methods are called, in order
#//
#//|
#//|   sub_seq.pre_start()         (task)
#//|   parent_seq.pre_do(0)        (task)
#//|   parent_req.mid_do(sub_seq)  (func)
#//|     sub_seq.body()            (task)
#//|   parent_seq.post_do(sub_seq) (func)
#//|   sub_seq.post_start()        (task)
#//|
#//
#// Remember, it is the *parent* sequence's pre|mid|post_do that are called, not
#// the sequence being executed.
#//
#//
#// Executing sequence items via <start_item>/<finish_item> or `uvm_do macros:
#//
#// Items are started in the <body> of a parent sequence via calls to
#// <start_item>/<finish_item> or invocations of any of the `uvm_do
#// macros. The <pre_do>, <mid_do>, and <post_do> methods of the parent
#// sequence will be called as the item is executed.
#//
#// The sequence-item execution flow looks like
#//
#// User code
#//
#//| parent_seq.start_item(item, priority)
#//| item.randomize(...) [with {constraints}]
#//| parent_seq.finish_item(item)
#//|
#//| or
#//|
#//| `uvm_do_with_prior(item, constraints, priority)
#//|
#//
#// The following methods are called, in order
#//
#//|
#//|   sequencer.wait_for_grant(prior) (task) \ start_item  \
#//|   parent_seq.pre_do(1)            (task) /              \
#//|                                                      `uvm_do* macros
#//|   parent_seq.mid_do(item)         (func) \              /
#//|   sequencer.send_request(item)    (func)  \finish_item /
#//|   sequencer.wait_for_item_done()  (task)  /
#//|   parent_seq.post_do(item)        (func) /
#//
#// Attempting to execute a sequence via <start_item>/<finish_item>
#// will produce a run-time error.
#//------------------------------------------------------------------------------


class UVMSequenceBase(UVMSequenceItem):

    #  // Variable: do_not_randomize
    #  //
    #  // If set, prevents the sequence from being randomized before being executed
    #  // by the `uvm_do*() and `uvm_rand_send*() macros,
    #  // or as a default sequence.
    #  //
    #  bit do_not_randomize


    #  static string type_name = "uvm_sequence_base"
    type_name = "uvm_sequence_base"

    def __init__(self, name="uvm_sequence"):
        """
          Function: __init__

          The constructor for uvm_sequence_base.

        Args:
            name:
        """
        UVMSequenceItem.__init__(self, name)
        self.m_sequence_state = UVM_CREATED
        self.m_wait_for_grant_semaphore = 0
        self.m_init_phase_daps(1)
        self.m_next_transaction_id = 1
        self.m_priority = -1
        self.m_tr_recorder = None  # uvm_recorder
        #self.m_automatic_phase_objection_dap = None  # uvm_get_to_lock_dap#(bit)
        #self.m_starting_phase_dap = None  # uvm_get_to_lock_dap#(uvm_phase)
        # Each sequencer will assign a sequence id.  When a sequence is talking to multiple
        # sequencers, each sequence_id is managed separately
        self.m_sqr_seq_ids = UVMPool()
        self.children_array = {}  # bit[uvm_sequence_base]
        self.response_queue = SeqItemQueue()
        self.response_queue_depth = 8
        self.response_queue_error_report_disabled = False
        #  bits to detect if is_relevant()/wait_for_relevant() are implemented
        self.is_rel_default = False
        self.wait_rel_default = False
        self.m_sequence_process = None  # process
        self.m_use_response_handler = False
        #self.m_resp_queue_event = UVMEvent("resp_queue_event")
        self.m_resp_queue_event = Event("resp_queue_event")
        self.m_resp_queue_event.clear()
        self.m_events: Dict[int, Event] = {}
        self.m_events[UVM_FINISHED] = Event("UVM_FINISHED")

    def is_item(self) -> bool:
        """
          Function: is_item

          Returns 1 on items and 0 on sequences. As this object is a sequence,
          `is_item` will always return 0.

        Returns:
        """
        return False


    def get_sequence_state(self):
        """
          Function: get_sequence_state

          Returns the sequence state as an enumerated value. Can use to wait on
          the sequence reaching or changing from one or more states.

        .. code-block:: python

         | wait(get_sequence_state() & (UVM_STOPPED|UVM_FINISHED))

        Returns:
        """
        return self.m_sequence_state

    async def wait_for_sequence_state(self, state_mask):
        """
          Task: wait_for_sequence_state

          Waits until the sequence reaches one of the given `state`. If the sequence
          is already in one of the state, this method returns immediately.

        .. code-block:: python

         | wait_for_sequence_state(UVM_STOPPED|UVM_FINISHED)
        Args:
            state_mask:
        Raises:
        """
        state = self.m_sequence_state & state_mask
        if state in self.m_events:
            await self.m_events[state].wait()
        else:
            raise Exception("Cannot wait for state " + str(state)
                    + " - not implemented")

    def get_tr_handle(self):
        """
          Function: get_tr_handle

          Returns the integral recording transaction handle for this sequence.
          Can be used to associate sub-sequences and sequence items as
          child transactions when calling <uvm_component::begin_child_tr>.
        Returns:
        """
        if (self.m_tr_recorder is not None):
            return self.m_tr_recorder.get_handle()
        else:
            return 0


    async def start(self, sequencer, parent_sequence=None, this_priority=-1,
            call_pre_post=1):
        """
         --------------------------
          Group: Sequence Execution
         --------------------------


          Task: start

          Executes this sequence, returning when the sequence has completed.

          The `sequencer` argument specifies the sequencer on which to run this
          sequence. The sequencer must be compatible with the sequence.

          If `parent_sequence` is `None`, then this sequence is a root parent,
          otherwise it is a child of `parent_sequence`. The `parent_sequence`'s
          pre_do, mid_do, and post_do methods will be called during the execution
          of this sequence.

          By default, the `priority` of a sequence
          is the priority of its parent sequence.
          If it is a root sequence, its default priority is 100.
          A different priority may be specified by `this_priority`.
          Higher numbers indicate higher priority.

          If `call_pre_post` is set to 1 (default), then the `pre_body` and
          `post_body` tasks will be called before and after the sequence
          `body` is called.

         async def start(self, uvm_sequencer_base sequencer,
                             uvm_sequence_base parent_sequence = None,
                             int this_priority = -1,
                             bit call_pre_post = 1)
        Args:
            sequencer:
            parent_sequence:
            this_priority:
            call_pre_post:
        """
        old_automatic_phase_objection = False
        self.set_item_context(parent_sequence, sequencer)

        if not(self.m_sequence_state in [UVM_CREATED,UVM_STOPPED,UVM_FINISHED]):
            uvm_fatal("SEQ_NOT_DONE",
                "Sequence " + self.get_full_name() + " already started")
        #end

        if self.m_parent_sequence is not None:
            self.m_parent_sequence.children_array[self] = 1

        if this_priority < -1:
            uvm_fatal("SEQPRI", sv.sformatf("Sequence %s start has illegal priority: %0d",
               self.get_full_name(), this_priority))

        if this_priority < 0:
            if parent_sequence is None:
                this_priority = 100
            else:
                this_priority = parent_sequence.get_priority()

        # Check that the response queue is empty from earlier runs
        self.clear_response_queue()
        self.m_priority = this_priority

        if self.m_sequencer is not None:
            handle = 0
            stream = None  # uvm_tr_stream stream
            if self.m_parent_sequence is None:
                stream = self.m_sequencer.get_tr_stream(self.get_name(), "Transactions")
                handle = self.m_sequencer.begin_tr(self, self.get_name())
                self.m_tr_recorder = UVMRecorder.get_recorder_from_handle(handle)
            else:
                stream = self.m_sequencer.get_tr_stream(self.get_root_sequence_name(), "Transactions")
                val = 0
                if self.m_parent_sequence.m_tr_recorder is not None:
                    val = self.m_parent_sequence.m_tr_recorder.get_handle()
                handle = self.m_sequencer.begin_child_tr(self, val, self.get_root_sequence_name())
                self.m_tr_recorder = UVMRecorder.get_recorder_from_handle(handle)

        # Ensure that the sequence_id is intialized in case this sequence has been stopped previously
        self.set_sequence_id(-1)
        # Remove all sqr_seq_ids
        self.m_sqr_seq_ids.delete()

        # Register the sequence with the sequencer if defined.
        if self.m_sequencer is not None:
            self.m_sequencer.m_register_sequence(self)

        # Change the state to PRE_START, do this before the fork so that
        # the "if (!(m_sequence_state inside {...}" works
        self.m_sequence_state = UVM_PRE_START

        self.m_sequence_process = cocotb.fork(self.start_process(parent_sequence, call_pre_post))
        await self.m_sequence_process

        if self.m_sequencer is not None:
            self.m_sequencer.end_tr(self)

        # Clean up any sequencer queues after exiting; if we
        # were forcibly stoped, this step has already taken place
        if self.m_sequence_state != UVM_STOPPED:
            if self.m_sequencer is not None:
                self.m_sequencer.m_sequence_exiting(self)

        #0; // allow stopped and finish waiters to resume
        await uvm_zero_delay()

        if (self.m_parent_sequence is not None and
                self in self.m_parent_sequence.children_array):
            del self.m_parent_sequence.children_array[self]

        old_automatic_phase_objection = self.get_automatic_phase_objection()
        self.m_init_phase_daps(1)
        self.set_automatic_phase_objection(old_automatic_phase_objection)


    async def start_process(self, parent_sequence, call_pre_post):
        #fork
        #m_sequence_process = process::self()

        # absorb delta to ensure PRE_START was seen
        #0
        await uvm_zero_delay()

        # Raise the objection if enabled
        # (This will lock the uvm_get_to_lock_dap)
        if self.get_automatic_phase_objection():
            self.m_safe_raise_starting_phase("automatic phase objection")

        await self.pre_start()

        if call_pre_post == 1:
            self.m_sequence_state = UVM_PRE_BODY
            #0
            await uvm_zero_delay()
            await self.pre_body()

        if parent_sequence is not None:
            parent_sequence.pre_do(0)    # task
            parent_sequence.mid_do(self)  # function

        self.m_sequence_state = UVM_BODY
        #0
        await uvm_zero_delay()
        await self.body()

        self.m_sequence_state = UVM_ENDED
        #0
        await uvm_zero_delay()

        if parent_sequence is not None:
            parent_sequence.post_do(self)

        if call_pre_post == 1:
            self.m_sequence_state = UVM_POST_BODY
            #0
            await uvm_zero_delay()
            await self.post_body()

        self.m_sequence_state = UVM_POST_START
        #0
        await uvm_zero_delay()
        await self.post_start()

        # Drop the objection if enabled
        if self.get_automatic_phase_objection():
            self.m_safe_drop_starting_phase("automatic phase objection")

        self.m_sequence_state = UVM_FINISHED
        self.m_events[UVM_FINISHED].set()
        #0
        await uvm_zero_delay()
        #join


    async def pre_start(self):
        """
          Task: pre_start

          This task is a user-definable callback that is called before the
          optional execution of `pre_body`.
          This method should not be called directly by the user.

         virtual task pre_start()
        """
        return

    async def pre_body(self):
        """
        Task: pre_body

        This task is a user-definable callback that is called before the
        execution of `body` `only` when the sequence is started with `start`.
        If `start` is called with `call_pre_post` set to 0, `pre_body` is not
        called.
        This method should not be called directly by the user.
        """
        return

    def pre_do(self, is_item):
        """
          Task: pre_do

          This task is a user-definable callback task that is called ~on the
          parent sequence~, if any
          sequence has issued a wait_for_grant() call and after the sequencer has
          selected this sequence, and before the item is randomized.

          Although pre_do is a task, consuming simulation cycles may result in
          unexpected behavior on the driver.

          This method should not be called directly by the user.
        Args:
            is_item:
        """
        return


    def mid_do(self, this_item):
        """
          Function: mid_do

          This function is a user-definable callback function that is called after
          the sequence item has been randomized, and just before the item is sent
          to the driver.  This method should not be called directly by the user.

        Args:
            this_item:
        """
        return

    async def body(self):
        """
          Task: body

          This is the user-defined task where the main sequence code resides.
          This method should not be called directly by the user.
        """
        uvm_warning("uvm_sequence_base", "Body definition undefined")
        await uvm_zero_delay()

    def post_do(self, this_item):
        """
          Function: post_do

          This function is a user-definable callback function that is called after
          the driver has indicated that it has completed the item, using either
          this item_done or put methods. This method should not be called directly
          by the user.
        Args:
            this_item:
        """
        return

    async def post_body(self):
        """
          Task: post_body

          This task is a user-definable callback task that is called after the
          execution of `body` `only` when the sequence is started with `start`.
          If `start` is called with `call_pre_post` set to 0, `post_body` is not
          called.
          This task is a user-definable callback task that is called after the
          execution of the body, unless the sequence is started with call_pre_post=0.
          This method should not be called directly by the user.
        """
        return

    async def post_start(self):
        """

          Task: post_start

          This task is a user-definable callback that is called after the
          optional execution of `post_body`.
          This method should not be called directly by the user.
        """
        return

    #  // Group: Run-Time Phasing
    #  //
    #
    #  // Automatic Phase Objection DAP
    #  local uvm_get_to_lock_dap#(bit) m_automatic_phase_objection_dap
    #  // Starting Phase DAP
    #  local uvm_get_to_lock_dap#(uvm_phase) m_starting_phase_dap

    def m_init_phase_daps(self, create):
        """
          Function- m_init_phase_daps
          Either creates or renames DAPS
        Args:
            create:
        """
        apo_name = sv.sformatf("%s.automatic_phase_objection", self.get_full_name())
        sp_name = sv.sformatf("%s.starting_phase", self.get_full_name())

        if create:
            self.m_automatic_phase_objection_dap = uvm_get_to_lock_dap.type_id.create(
                apo_name, self.get_sequencer())
            self.m_starting_phase_dap = uvm_get_to_lock_dap.type_id.create(sp_name,
                    self.get_sequencer())
        else:
            self.m_automatic_phase_objection_dap.set_name(apo_name)
            self.m_starting_phase_dap.set_name(sp_name)


    def get_starting_phase(self):
        """
          Function: get_starting_phase
          Returns the 'starting phase'.

          If non-`None`, the starting phase specifies the phase in which this
          sequence was started.  The starting phase is set automatically when
          this sequence is started as the default sequence on a sequencer.
          See <uvm_sequencer_base::start_phase_sequence> for more information.

          Internally, the `uvm_sequence_base` uses an `uvm_get_to_lock_dap` to
          protect the starting phase value from being modified
          after the reference has been read.  Once the sequence has ended
          its execution (either via natural termination, or being killed),
          then the starting phase value can be modified again.

        Returns:
        """
        return self.m_starting_phase_dap.get()

    def set_starting_phase(self, phase):
        """

          Function: set_starting_phase
          Sets the 'starting phase'.

          Internally, the `uvm_sequence_base` uses a `uvm_get_to_lock_dap` to
          protect the starting phase value from being modified
          after the reference has been read.  Once the sequence has ended
          its execution (either via natural termination, or being killed),
          then the starting phase value can be modified again.

        Args:
            phase:
        """
        self.m_starting_phase_dap.set(phase)

    def set_automatic_phase_objection(self, value):
        """
          Function: set_automatic_phase_objection
          Sets the 'automatically object to starting phase' bit.

          The most common interaction with the starting phase
          within a sequence is to simply `raise` the phase's objection
          prior to executing the sequence, and `drop` the objection
          after ending the sequence (either naturally, or
          via a call to `kill`). In order to
          simplify this interaction for the user, the UVM
          provides the ability to perform this functionality
          automatically.

          For example:
        .. code-block:: python

         | function my_sequence::new(string name="unnamed")
         |   super.new(name)
         |   set_automatic_phase_objection(1)
         | endfunction : new

          From a timeline point of view, the automatic phase objection
          looks like:
        .. code-block:: python

         | start() is executed
         |   --! Objection is raised !--
         |   pre_start() is executed
         |   pre_body() is optionally executed
         |   body() is executed
         |   post_body() is optionally executed
         |   post_start() is executed
         |   --! Objection is dropped !--
         | start() unblocks

          This functionality can also be enabled in sequences
          which were not written with UVM Run-Time Phasing in mind:
        .. code-block:: python

         | my_legacy_seq_type seq = new("seq")
         | seq.set_automatic_phase_objection(1)
         | seq.start(my_sequencer)

          Internally, the `uvm_sequence_base` uses a `uvm_get_to_lock_dap` to
          protect the `automatic_phase_objection` value from being modified
          after the reference has been read.  Once the sequence has ended
          its execution (either via natural termination, or being killed),
          then the `automatic_phase_objection` value can be modified again.

          NEVER set the automatic phase objection bit to 1 if your sequence
          runs with a forever loop inside of the body, as the objection will
          never get dropped!
        Args:
            value:
        """
        self.m_automatic_phase_objection_dap.set(value)

    def get_automatic_phase_objection(self):
        """
          Function: get_automatic_phase_objection
          Returns (and locks) the value of the 'automatically object to
          starting phase' bit.

          If 1, then the sequence will automatically raise an objection
          to the starting phase (if the starting phase is not `None`) immediately
          prior to `pre_start` being called.  The objection will be dropped
          after `post_start` has executed, or `kill` has been called.

        Returns:
        """
        return self.m_automatic_phase_objection_dap.get()

    def m_safe_raise_starting_phase(self, description="", count=1):
        """
          m_safe_raise_starting_phase
         function void m_safe_raise_starting_phase(string description = "",
                                                   int count = 1)
        Args:
            description:
            count:
        """
        starting_phase = self.get_starting_phase()
        if (starting_phase is not None):
            starting_phase.raise_objection(self, description, count)

    def m_safe_drop_starting_phase(self, description="", count=1):
        """

          m_safe_drop_starting_phase
         function void m_safe_drop_starting_phase(string description = "",
                                                  int count = 1)
        Args:
            description:
            count:
        """
        starting_phase = self.get_starting_phase()
        if (starting_phase is not None):
            starting_phase.drop_objection(self, description, count)
        #  endfunction : m_safe_drop_starting_phase

    def set_priority(self, value):
        """
         ------------------------
          Group: Sequence Control
         ------------------------

          Function: set_priority

          The priority of a sequence may be changed at any point in time.  When the
          priority of a sequence is changed, the new priority will be used by the
          sequencer the next time that it arbitrates between sequences.

          The default priority value for a sequence is 100.  Higher values result
          in higher priorities.
        Args:
            value:
        """
        self.m_priority = value

    def get_priority(self):
        """
          Function: get_priority

          This function returns the current priority of the sequence.

        Returns:
        """
        return self.m_priority

    def is_relevant(self):
        """
          Function: is_relevant

          The default is_relevant implementation returns 1, indicating that the
          sequence is always relevant.

          Users may choose to override with their own virtual function to indicate
          to the sequencer that the sequence is not currently relevant after a
          request has been made.

          When the sequencer arbitrates, it will call is_relevant on each requesting,
          unblocked sequence to see if it is relevant. If a 0 is returned, then the
          sequence will not be chosen.

          If all requesting sequences are not relevant, then the sequencer will call
          wait_for_relevant on all sequences and re-arbitrate upon its return.

          Any sequence that implements is_relevant must also implement
          wait_for_relevant so that the sequencer has a way to wait for a
          sequence to become relevant.

        Returns:
        """
        self.is_rel_default = True
        return 1

    async def wait_for_relevant(self):
        """
          Task: wait_for_relevant

          This method is called by the sequencer when all available sequences are
          not relevant.  When wait_for_relevant returns the sequencer attempt to
          re-arbitrate.

          Returning from this call does not guarantee a sequence is relevant,
          although that would be the ideal. The method provide some delay to
          prevent an infinite loop.

          If a sequence defines is_relevant so that it is not always relevant (by
          default, a sequence is always relevant), then the sequence must also supply
          a wait_for_relevant method.
        """
        e = Event('e')
        wait_rel_default = True
        if self.is_rel_default != wait_rel_default:
            uvm_fatal("RELMSM",
                "is_relevant() was implemented without defining wait_for_relevant()")
        await e.wait()  # this is intended to never return


    #  // Task: lock
    #  //
    #  // Requests a lock on the specified sequencer. If sequencer is ~None~, the lock
    #  // will be requested on the current default sequencer.
    #  //
    #  // A lock request will be arbitrated the same as any other request.  A lock is
    #  // granted after all earlier requests are completed and no other locks or
    #  // grabs are blocking this sequence.
    #  //
    #  // The lock call will return when the lock has been granted.
    #
    #  task lock(uvm_sequencer_base sequencer = None)
    #    if (sequencer == None)
    #      sequencer = self.m_sequencer
    #
    #    if (sequencer == None)
    #      uvm_report_fatal("LOCKSEQR", "None self.m_sequencer reference", UVM_NONE)
    #
    #    sequencer.lock(this)
    #  endtask



    #  // Task: grab
    #  //
    #  // Requests a lock on the specified sequencer.  If no argument is supplied,
    #  // the lock will be requested on the current default sequencer.
    #  //
    #  // A grab request is put in front of the arbitration queue. It will be
    #  // arbitrated before any other requests. A grab is granted when no other grabs
    #  // or locks are blocking this sequence.
    #  //
    #  // The grab call will return when the grab has been granted.
    #
    #  task grab(uvm_sequencer_base sequencer = None)
    #    if (sequencer == None) begin
    #      if (self.m_sequencer == None) begin
    #        uvm_report_fatal("GRAB", "None self.m_sequencer reference", UVM_NONE)
    #      end
    #      self.m_sequencer.grab(this)
    #    end
    #    else begin
    #      sequencer.grab(this)
    #    end
    #  endtask


    #  // Function: unlock
    #  //
    #  // Removes any locks or grabs obtained by this sequence on the specified
    #  // sequencer. If sequencer is ~None~, then the unlock will be done on the
    #  // current default sequencer.
    #
    #  function void  unlock(uvm_sequencer_base sequencer = None)
    #    if (sequencer == None) begin
    #      if (self.m_sequencer == None) begin
    #        uvm_report_fatal("UNLOCK", "None self.m_sequencer reference", UVM_NONE)
    #      end
    #      self.m_sequencer.unlock(this)
    #    end else begin
    #      sequencer.unlock(this)
    #    end
    #  endfunction


    #  // Function: ungrab
    #  //
    #  // Removes any locks or grabs obtained by this sequence on the specified
    #  // sequencer. If sequencer is ~None~, then the unlock will be done on the
    #  // current default sequencer.
    #
    #  function void  ungrab(uvm_sequencer_base sequencer = None)
    #    unlock(sequencer)
    #  endfunction


    def is_blocked(self):
        """
          Function: is_blocked

          Returns a bit indicating whether this sequence is currently prevented from
          running due to another lock or grab. A 1 is returned if the sequence is
          currently blocked. A 0 is returned if no lock or grab prevents this
          sequence from executing. Note that even if a sequence is not blocked, it
          is possible for another sequence to issue a lock or grab before this
          sequence can issue a request.

        Returns:
        """
        return self.m_sequencer.is_blocked(self)


    def has_lock(self):
        """
          Function: has_lock

          Returns 1 if this sequence has a lock, 0 otherwise.

          Note that even if this sequence has a lock, a child sequence may also have
          a lock, in which case the sequence is still blocked from issuing
          operations on the sequencer.
        Returns:
        """
        return self.m_sequencer.has_lock(self)


    #  // Function: kill
    #  //
    #  // This function will kill the sequence, and cause all current locks and
    #  // requests in the sequence's default sequencer to be removed. The sequence
    #  // state will change to UVM_STOPPED, and the post_body() and post_start() callback
    #  // methods will not be executed.
    #  //
    #  // If a sequence has issued locks, grabs, or requests on sequencers other than
    #  // the default sequencer, then care must be taken to unregister the sequence
    #  // with the other sequencer(s) using the sequencer unregister_sequence()
    #  // method.
    #
    #  function void kill()
    #    if (m_sequence_process is not None) begin
    #      // If we are not connected to a sequencer, then issue
    #      // kill locally.
    #      if (self.m_sequencer == None) begin
    #        m_kill()
    #        // We need to drop the objection if we raised it...
    #        if (get_automatic_phase_objection()) begin
    #           m_safe_drop_starting_phase("automatic phase objection")
    #        end
    #        return
    #      end
    #      // If we are attached to a sequencer, then the sequencer
    #      // will clear out queues, and then kill this sequence
    #      self.m_sequencer.kill_sequence(this)
    #      // We need to drop the objection if we raised it...
    #      if (get_automatic_phase_objection()) begin
    #         m_safe_drop_starting_phase("automatic phase objection")
    #      end
    #      return
    #    end
    #  endfunction


    def do_kill(self):
        """
          Function: do_kill

          This function is a user hook that is called whenever a sequence is
          terminated by using either sequence.kill() or sequencer.stop_sequences()
          (which effectively calls sequence.kill()).
        """
        return


    #  function void m_kill()
    #    do_kill()
    #    foreach(children_array[i]) begin
    #       i.kill()
    #    end
    #    if (m_sequence_process is not None) begin
    #      m_sequence_process.kill
    #      m_sequence_process = None
    #    end
    #    m_sequence_state = UVM_STOPPED
    #    if ((self.m_parent_sequence is not None) && (self.m_parent_sequence.children_array.exists(this)))
    #      self.m_parent_sequence.children_array.delete(this)
    #  endfunction


    #  //-------------------------------
    #  // Group: Sequence Item Execution
    #  //-------------------------------


    def create_item(self, type_var, l_sequencer, name):
        """
          Function: create_item

          Create_item will create and initialize a sequence_item or sequence
          using the factory.  The sequence_item or sequence will be initialized
          to communicate with the specified sequencer.

         protected function uvm_sequence_item create_item(uvm_object_wrapper type_var,
                                                          uvm_sequencer_base l_sequencer, string name)
        Args:
            type_var:
            l_sequencer:
            name:
        Returns:
        """
        from ..base.uvm_coreservice import UVMCoreService
        cs = UVMCoreService.get()
        factory = cs.get_factory()
        item = factory.create_object_by_type(type_var, self.get_full_name(), name)
        item.set_item_context(self, l_sequencer)
        return item


    async def start_item(self, item, set_priority=-1, sequencer=None):
        """
        `start_item` and `finish_item` together will initiate operation of
        a sequence item.  If the item has not already been
        initialized using create_item, then it will be initialized here to use
        the default sequencer specified by self.m_sequencer.  Randomization
        may be done between start_item and finish_item to ensure late generation

         virtual task start_item (uvm_sequence_item item,
                                  int set_priority = -1,
                                  uvm_sequencer_base sequencer=None)
        Args:
            item:
            set_priority:
            sequencer:
        """
        seq = None

        if item is None:
            uvm_fatal("NoneITM",
               "attempting to start a None item from sequence " + self. get_full_name())
            return

        #if($cast(seq, item)) begin
        #  uvm_report_fatal("SEQNOTITM",
        #     {"attempting to start a sequence using start_item() from sequence '",
        #      get_full_name(), "'. Use seq.start() instead."}, UVM_NONE)
        #  return
        seq = item

        if sequencer is None:
            sequencer = item.get_sequencer()

        if sequencer is None:
            sequencer = self.get_sequencer()

        if sequencer is None:
            uvm_fatal("SEQ", SEQ_ERR1_MSG + self.get_full_name())
            return

        item.set_item_context(self, sequencer)

        if set_priority < 0:
            set_priority = self.get_priority()

        await sequencer.wait_for_grant(self, set_priority)

        # TODO recording
        #if (sequencer.is_auto_item_recording_enabled()) begin
        #  void'(sequencer.begin_child_tr(item,
        #                                 (self.m_tr_recorder == None) ? 0 : self.m_tr_recorder.get_handle(),
        #                                 item.get_root_sequence_name(), "Transactions"))
        #end
        self.pre_do(1)

    async def finish_item(self, item, set_priority=-1):
        """
        finish_item, together with start_item together will initiate operation of
        a sequence_item.  Finish_item must be called
        after start_item with no delays or delta-cycles.  Randomization, or other
        functions may be called between the start_item and finish_item calls.

         virtual task finish_item (uvm_sequence_item item,
                                   int set_priority = -1)
        Args:
            item (UVMSequenceItem): Item that was started with start_item().
            set_priority (int): Priority for the item.
        """
        sequencer = item.get_sequencer()

        if sequencer is None:
            uvm_fatal("STRITM", "sequence_item has None sequencer")

        self.mid_do(item)
        sequencer.send_request(self, item)
        await sequencer.wait_for_item_done(self, -1)

        # tpoikela: commented out, otherwise item.end_event is not triggered
        # if sequencer.is_auto_item_recording_enabled():
        sequencer.end_tr(item)
        self.post_do(item)

    #  // Task: wait_for_grant
    #  //
    #  // This task issues a request to the current sequencer.  If item_priority is
    #  // not specified, then the current sequence priority will be used by the
    #  // arbiter. If a lock_request is made, then the sequencer will issue a lock
    #  // immediately before granting the sequence.  (Note that the lock may be
    #  // granted without the sequence being granted if is_relevant is not asserted).
    #  //
    #  // When this method returns, the sequencer has granted the sequence, and the
    #  // sequence must call send_request without inserting any simulation delay
    #  // other than delta cycles.  The driver is currently waiting for the next
    #  // item to be sent via the send_request call.
    #
    #  virtual task wait_for_grant(int item_priority = -1, bit lock_request = 0)
    #    if (self.m_sequencer == None) begin
    #      uvm_fatal("WAITGRANT", "None self.m_sequencer reference")
    #    end
    #    self.m_sequencer.wait_for_grant(this, item_priority, lock_request)
    #  endtask


    #  // Function: send_request
    #  //
    #  // The send_request function may only be called after a wait_for_grant call.
    #  // This call will send the request item to the sequencer, which will forward
    #  // it to the driver. If the rerandomize bit is set, the item will be
    #  // randomized before being sent to the driver.
    #
    #  virtual function void send_request(uvm_sequence_item request, bit rerandomize = 0)
    #    if (self.m_sequencer == None) begin
    #        uvm_report_fatal("SENDREQ", "None self.m_sequencer reference", UVM_NONE)
    #      end
    #    self.m_sequencer.send_request(this, request, rerandomize)
    #  endfunction


    #  // Task: wait_for_item_done
    #  //
    #  // A sequence may optionally call wait_for_item_done.  This task will block
    #  // until the driver calls item_done or put.  If no transaction_id parameter
    #  // is specified, then the call will return the next time that the driver calls
    #  // item_done or put.  If a specific transaction_id is specified, then the call
    #  // will return when the driver indicates completion of that specific item.
    #  //
    #  // Note that if a specific transaction_id has been specified, and the driver
    #  // has already issued an item_done or put for that transaction, then the call
    #  // will hang, having missed the earlier notification.
    #
    #
    #  virtual task wait_for_item_done(int transaction_id = -1)
    #    if (self.m_sequencer == None) begin
    #        uvm_report_fatal("WAITITEMDONE", "None self.m_sequencer reference", UVM_NONE)
    #      end
    #    self.m_sequencer.wait_for_item_done(this, transaction_id)
    #  endtask



    def use_response_handler(self, enable):
        """
          Group: Response API
         --------------------

          Function: use_response_handler

          When called with enable set to 1, responses will be sent to the response
          handler. Otherwise, responses must be retrieved using get_response.

          By default, responses from the driver are retrieved in the sequence by
          calling get_response.

          An alternative method is for the sequencer to call the response_handler
          function with each response.
        Args:
            enable:
        """
        self.m_use_response_handler = enable


    def get_use_response_handler(self):
        """
          Function: get_use_response_handler

          Returns the state of the use_response_handler bit.

        Returns:
        """
        return self.m_use_response_handler

    def response_handler(self, response):
        """
          Function: response_handler

          When the use_response_handler bit is set to 1, this virtual task is called
          by the sequencer for each response that arrives for this sequence.

        Args:
            response:
        """
        return


    def set_response_queue_error_report_disabled(self, value):
        """
          Function: set_response_queue_error_report_disabled

          By default, if the self.response_queue overflows, an error is reported. The
          self.response_queue will overflow if more responses are sent to this sequence
          from the driver than get_response calls are made. Setting value to 0
          disables these errors, while setting it to 1 enables them.

        Args:
            value:
        """
        self.response_queue_error_report_disabled = value

    def get_response_queue_error_report_disabled(self):
        """
          Function: get_response_queue_error_report_disabled

          When this bit is 0 (default value), error reports are generated when
          the response queue overflows. When this bit is 1, no such error
          reports are generated.

        Returns:
        """
        return self.response_queue_error_report_disabled

    def set_response_queue_depth(self, value):
        """
          Function: set_response_queue_depth

          The default maximum depth of the response queue is 8. These method is used
          to examine or change the maximum depth of the response queue.

          Setting the response_queue_depth to -1 indicates an arbitrarily deep
          response queue.  No checking is done.

        Args:
            value:
        """
        self.response_queue_depth = value

    def get_response_queue_depth(self):
        """
          Function: get_response_queue_depth

          Returns the current depth setting for the response queue.

        Returns:
        """
        return self.response_queue_depth

    def clear_response_queue(self):
        """
          Function: clear_response_queue

          Empties the response queue for this sequence.

        """
        self.response_queue.delete()
        self.m_resp_queue_event.set()

    def put_base_response(self, response):
        """
         virtual function void put_base_response(input uvm_sequence_item response)
        Args:
            response:
        """
        if ((self.response_queue_depth == -1) or
                (self.response_queue.size() < self.response_queue_depth)):
            self.response_queue.push_back(response)
            self.m_resp_queue_event.set()
            return
        if self.response_queue_error_report_disabled == 0:
            uvm_error(self.get_full_name(), "Response queue overflow, response was dropped")

    def put_response(self, response_item):
        """
          Function- put_response

          Internal method.
         virtual function void put_response (uvm_sequence_item response_item)
        Args:
            response_item:
        """
        self.put_base_response(response_item)

    async def get_base_response(self, response: SeqItemQueue, transaction_id=-1):
        """
          Function- get_base_response
         virtual task get_base_response(output uvm_sequence_item response, input int transaction_id = -1)
        Args:
            response:
            transaction_id:
        """
        queue_size = 0
        i = 0

        if self.response_queue.size() == 0:
            await wait(lambda: self.response_queue.size() != 0,
                 self.m_resp_queue_event)

        if transaction_id == -1:
            resp_item = self.response_queue.pop_front()
            response.append(resp_item)
            return

        while True:
            queue_size = self.response_queue.size()
            for i in range(queue_size):
                seq_item = self.response_queue.get(i)
                if (seq_item.get_transaction_id() == transaction_id):
                    response.append(seq_item)
                    self.response_queue.delete(i)
                    self.m_resp_queue_event.set()
                    return
            #wait (self.response_queue.size() != queue_size)
            await wait(lambda: self.response_queue.size() != queue_size,
                       self.m_resp_queue_event)


    #  //----------------------
    #  // Misc Internal methods
    #  //----------------------

    def m_get_sqr_sequence_id(self, sequencer_id, update_sequence_id):
        """
          m_get_sqr_sequence_id
          ---------------------
         function int m_get_sqr_sequence_id(int sequencer_id, bit update_sequence_id)
        Args:
            sequencer_id:
            update_sequence_id:
        Returns:
        """
        if self.m_sqr_seq_ids.exists(sequencer_id):
            if update_sequence_id == 1:
                self.set_sequence_id(self.m_sqr_seq_ids[sequencer_id])
            return self.m_sqr_seq_ids[sequencer_id]

        if update_sequence_id == 1:
            self.set_sequence_id(-1)
        return -1


    def m_set_sqr_sequence_id(self, sequencer_id, sequence_id):
        """
          m_set_sqr_sequence_id
          ---------------------
         function void m_set_sqr_sequence_id(int sequencer_id, int sequence_id)
        Args:
            sequencer_id:
            sequence_id:
        """
        self.m_sqr_seq_ids[sequencer_id] = sequence_id
        self.set_sequence_id(sequence_id)
